import os
import sys

import dbus
from dbus.mainloop.qt import DBusQtMainLoop

from PyQt4 import QtDeclarative
from PyQt4.QtCore import QUrl, QObject, QSettings
from PyQt4.QtGui import QApplication#, QIcon

from totallywired.utils.caching import init_caching
from totallywired.utils.services import check_tw_services_are_running
from totallywired.models.music import MusicCollection
from totallywired.models.indexes import IndexModel
from totallywired.models.queue import QueueModel
from totallywired.models.stations import RadioStationModel
from totallywired.player.client import QMusicPlayerClient
from totallywired.db.client import CollectionManager
from totallywired.utils.misc import MiscUtils

init_caching()


qtloop = DBusQtMainLoop()
SESSION_BUS = dbus.SessionBus(mainloop=qtloop)
check_tw_services_are_running(SESSION_BUS)


class TotallyWiredApplication(object):

    def __init__(self):
        self.app = QApplication(sys.argv)
        self.app.setOrganizationName('HomeBrewSoftware')
        self.app.setOrganizationDomain('homebrewsoftware.com')
        self.app.setApplicationName('Totally Wired')
        self.app.lastWindowClosed.connect(self._on_close)

        self._handle_argv()
        self.settings = QSettings(self.app)

        self.view = QtDeclarative.QDeclarativeView()
        title = self.ignored and self.view.tr("Totally Wired (ignored client)") or self.view.tr("Totally Wired")
        self.view.setWindowTitle(title)
        #~ view.setWindowIcon(QIcon())
        self.view.setResizeMode(QtDeclarative.QDeclarativeView.SizeRootObjectToView)

        self.utils = MiscUtils(self)
        collections = CollectionManager(SESSION_BUS, self.view, self.settings)
        queue = QueueModel(self, collections)
        music = MusicCollection(self, queue, collections)
        self.stations = RadioStationModel()
        self.player = QMusicPlayerClient(self, music, queue, SESSION_BUS)
        indexes = IndexModel(collections)
        music.reload()
        self.stations.update_stations()

        def ON_DATABASE_RELOADED():
            music.reload_model()
        def ON_DOCUMENT_CHANGED(docid):
            music.reload_document(docid)
            queue.reload_document(docid)
        def ON_COLLECTION_CHANGED(i, collection):
            indexes.collection_changed(i, collection)
            music._cdda.collection_changed(i, collection)
        def ON_TOPLEVEL_FOCUS_CHANGE(widget):
            if widget is None and self:
                return
            self.player.grab_mediakeys()

        collections.databaseReloaded.connect(ON_DATABASE_RELOADED)
        collections.documentChanged.connect(ON_DOCUMENT_CHANGED)
        collections.collectionChanged.connect(ON_COLLECTION_CHANGED)
        self.app.focusChanged.connect(ON_TOPLEVEL_FOCUS_CHANGE)

        rc = self.view.rootContext()
        rc.setContextProperty('Stations', self.stations)
        rc.setContextProperty('Player', self.player)
        rc.setContextProperty('DbClient', collections)
        rc.setContextProperty('Indexes', indexes)
        rc.setContextProperty('Collection', music)
        rc.setContextProperty('Queue', queue)
        rc.setContextProperty('Utils', self.utils)

        # load the main QML file into the view
        qmlpath = os.path.join(os.path.dirname(__file__), "qml/tw.qml")
        self.view.setSource(QUrl.fromLocalFile(qmlpath))
        return

    # private
    def _handle_argv(self):
        self.ignored = "--ignore" in sys.argv

        if "--fresh-config" in sys.argv:
            conf = os.path.expanduser("~/.config/HomeBrewSoftware/Totally Wired.conf")
            if os.path.exists(conf):
                try:
                    os.remove(conf)
                except Exception as e:
                    print("Failed to remove config file, reason: %s" % e)
        return

    def _store_session_state(self):
        settings = self.settings
        size = self.view.size()

        # mainwindow aspects
        settings.beginGroup("mainwindow")
        settings.setValue("width", size.width())
        settings.setValue("height", size.height())
        settings.setValue("fullScreen", self.view.isFullScreen())
        settings.endGroup()

        # player stuff
        settings.beginGroup("musicplayer")
        settings.setValue("shuffle", self.player.shuffle)
        settings.setValue("repeat", self.player.repeat)
        settings.endGroup()

        # qml aspects
        settings.beginGroup("qml")
        o = self.view.rootObject()
        for obj in ("musicLibraryView", "musicView", "queueView", "cddaView"):
            it = o.findChild(QObject, obj)
            if not it: continue
            it.saveState()

        settings.endGroup()
        # write any unsaved changes
        settings.sync()
        return

    def _on_close(self):
        if (not ("--keep-playing-on-exit" in sys.argv or
            self.ignored) and
            self.player.player is not None and
            self.player.is_playing):
            self.player.pause()

        if not self.ignored:
            self._store_session_state()
        return

    # public
    def run(self):
        self.view.show()
        sys.exit(self.app.exec_())
        return
