import dbus

import sys
import time
import xapian
import threading

from gi.repository import GLib
from operator import attrgetter
from PyQt4.QtCore import QObject, pyqtSlot, pyqtSignal, pyqtProperty

from totallywired.db.collections.client import (
    ClientSideCollection,
    ClientSideCddaCollection)

from totallywired.db.collectors import get_metadata
from totallywired.db.collectors.enums import DbEnums, Mappings
from totallywired.db.collectors.filenames import FilenameCollector
from totallywired.db.collectors.filesystem import FileSystemCollector

from totallywired.utils.paths import is_cdda_path
from totallywired.utils.misc import StdoutColors

from totallywired.utils.db import (
    get_collated_database,
    get_writable_twxdbs_for_uris,
    get_term_generator,
    get_doc_from_uri)


class LikeSwitchJob(threading.Thread):

    def __init__(self, dislike, uris, finished_cb):
        threading.Thread.__init__(self)
        self.dislike = dislike
        self.uris = uris
        self.finished_cb = finished_cb
        return

    def _new_doc(self, doc):
        _v = doc.get_value(DbEnums.V_LIKE)
        new_doc = xapian.Document(doc)

        # scrub old lkd terms and values
        if _v:
            new_doc.remove_value(DbEnums.V_LIKE)
        for term in doc.termlist():
            if term.term.startswith("LKD"):
                new_doc.remove_term(term.term)
        return new_doc, _v

    def _switch_like_value(self, db, uri):
        doc, docid = get_doc_from_uri(db, uri)
        if doc == -1:
            return

        # update document values and terms and replace in db
        new_doc, _v = self._new_doc(doc)
        if _v == "1":
            new_doc.add_value(DbEnums.V_LIKE, "0")
        else:
            new_doc.add_value(DbEnums.V_LIKE, "1")
            new_doc.add_term(DbEnums.T_LIKE)

        db.replace_document(docid, new_doc)
        return

    def _switch_dislike_value(self, db, uri):
        doc, docid = get_doc_from_uri(db, uri)
        if doc == -1:
            return

        # update document values and terms and replace in db
        new_doc, _v = self._new_doc(doc)
        if _v == "2":
            new_doc.add_value(DbEnums.V_LIKE, "0")
        else:
            new_doc.add_value(DbEnums.V_LIKE, "2")
            new_doc.add_term(DbEnums.T_DISLIKE)

        db.replace_document(docid, new_doc)
        return

    def run(self):
        dbs = get_writable_twxdbs_for_uris(self.uris)
        for src, (db, uris) in dbs.items():
            for uri in uris:
                # edit values
                if self.dislike:
                    self._switch_dislike_value(db, uri)
                else:
                    self._switch_like_value(db, uri)
            # save changes for each db
            db.commit()
            db.close()

        self.finished_cb()
        return


class _WaveInfo(object):

    def __init__(self, path):
        self.length = FilenameCollector.try_get_wave_length(path)
        return


class ChangeData(object):

    def __init__(self, changes, path):
        self.changes = changes
        self.path = path
        self.data, self.kmap = get_metadata(path)
        return

    @property
    def info(self):
        if self.data:
            return self.data.info
        return _WaveInfo(self.path)

    def __getitem__(self, k):
        if k in self.changes:
            return self.changes[k]
        return self.data[self.kmap[k]]

    def __contains__(self, k):
        return k in self.changes


class MetadataUpdateJob(threading.Thread):

    def __init__(self, uri_change_map, finished_cb, *args):
        threading.Thread.__init__(self)
        self.uri_change_map = uri_change_map
        self._finished_cb = finished_cb
        self._args = args
        return

    def run(self):
        uris = self.uri_change_map.keys()
        dbs = get_writable_twxdbs_for_uris(uris)
        i = 0
        for src, (db, uris) in dbs.items():
            for uri in uris:
                path = GLib.filename_from_uri(uri, "")
                #~ print("%s.UpdateJob %s terms:%s" % (i, path, self.uri_change_map[uri]))
                data = ChangeData(self.uri_change_map[uri], path)

                doc, docid = get_doc_from_uri(db, uri)
                v = doc.get_value(DbEnums.V_LIKE)

                if v == "1":    # liked
                    disliked = []
                    liked = [doc.get_data(),]
                elif v == "2":  # disliked
                    liked = []
                    disliked = [doc.get_data(),]
                else:
                    liked = []
                    disliked = []

                new_doc = xapian.Document()

                FileSystemCollector.index_document(
                    new_doc, get_term_generator(db),
                    path, 
                    data,   # metadata changes
                    Mappings.GENERIC,
                    liked, disliked)

                db.replace_document(docid, new_doc)
                print("%s.UpdateJob %s terms:%s" % (i, path, [term.term for term in doc.termlist()]))
                i += 1

            db.commit()
            db.close()

        # all done
        self._finished_cb(*self._args)
        return


class CollectionManager(QObject):

    # const
    DEFAULT_QUERY_FLAGS = (xapian.QueryParser.FLAG_PARTIAL|
                           xapian.QueryParser.FLAG_BOOLEAN|
                           xapian.QueryParser.FLAG_SPELLING_CORRECTION)

    SERVICE_COMM_ATTEMPT_COUNT = 20
    SERVICE_COMM_ATTEMPT_DELAY = 0.25

    AUTO_MOUNT_STARTUP = 0  # default
    AUTO_MOUNT_ALWAYS = 1
    AUTO_MOUNT_NEVER = 2

    # signals
    collectionChanged = pyqtSignal(int, object)
    collectionAdded = pyqtSignal(int, str)
    documentChanged = pyqtSignal(int)
    databaseReloaded = pyqtSignal()

    def __init__(self, bus, view, settings):
        QObject.__init__(self)
        self.view = view
        self.service = None
        self._collections = []

        attempt = 0
        self.service = None
        while attempt <= self.SERVICE_COMM_ATTEMPT_COUNT:
            try:
                self.service = bus.get_object(
                    'org.tw.XCollectionManagerService',
                    '/XCollectionManagerService')
                break
            except dbus.DBusException:
                attempt += 1
                print(StdoutColors.BOLD + StdoutColors.WARNING + "No service 'org.tw.XCollectionManagerService', retrying... (attempt %s/%s)" % (attempt, self.SERVICE_COMM_ATTEMPT_COUNT) + StdoutColors.ENDC)
                time.sleep(self.SERVICE_COMM_ATTEMPT_DELAY)

        if self.service is not None:
            bus.add_signal_receiver(
                self.collection_changed,
                dbus_interface='org.tw.XCollectionManagerService',
                signal_name='CollectionChanged')
            bus.add_signal_receiver(
                self.reload_database,
                dbus_interface='org.tw.XCollectionManagerService',
                signal_name='ReloadDatabase')
            bus.add_signal_receiver(
                self.chooser_response,
                dbus_interface='org.tw.XCollectionManagerService',
                signal_name='ChooserResponse')
        else:
            print("Service 'org.tw.XCollectionManagerService' could not be contacted!!")

        if "--automount-never" in sys.argv:
            self._automount = self.AUTO_MOUNT_NEVER
        elif "--automount-always" in sys.argv:
            self._automount = self.AUTO_MOUNT_ALWAYS
            # this will cause any call to 'reload_collections'
            # to also demand a mounting of any currently unmounted 
            # volume that contains a tw collection
        else:
            self._automount = self.AUTO_MOUNT_STARTUP
            self.service.auto_mount_collections()

        self.db = self.get_database()
        return

    def reload_collections(self):
        def _collection(c):
            if is_cdda_path(str(c['Source'])):
                return ClientSideCddaCollection(c)
            return ClientSideCollection(c)

        collections = []
        automount = self._automount == self.AUTO_MOUNT_ALWAYS

        for c in self.service.get_collections(automount):
            collections.append(_collection(c))
        collections = sorted(
            collections,
            key=attrgetter("removable", "live_sorter", "title"))
        self._collections = collections
        return

    def collection_changed(self, _dict):
        source = str(_dict["Source"])
        for i, c in enumerate(self._collections):
            if c.source == source:
                break
        else:
            return

        c_was_live = c.live
        c.from_dict(_dict)
        if c_was_live != c.live:
            self.reload_collections()

            # repeat with new order
            for j, c in enumerate(self._collections):
                if c.source == source:
                    break
        else:
            j = i

        for k in range(min(i,j), max(i,j)+1):
            self.collectionChanged.emit(k, c)
        return

    def list(self):
        return self._collections

    def get_index(self, source):
        i = [i_c for i_c in enumerate(self._collections) if i_c[1].source == source]
        if i:
            return i[0]
        return False

    def get_cdda_collection(self):
        c = [c for c in self._collections if c.is_cdda and c.live]
        if len(c) > 0:
            return c[0]
        return False

    @pyqtSlot()
    def display_folder_chooser(self):
        wid = self.view.winId()
        self.service.display_gtk_chooser_dialog(wid, "")
        return

    @pyqtProperty(bool, notify=collectionChanged)
    def cdda_present(self):
        c = [c for c in self._collections if c.is_cdda and c.live]
        return len(c) > 0

    @pyqtSlot(str)
    def toggle_hidden(self, source):
        self.service.toggle_hidden(str(source))
        return

    @pyqtSlot(str)
    def build_collection(self, source):
        return self.service.build_collection(str(source))

    @pyqtSlot(str)
    def trash_collection(self, collection_source):
        return self.service.trash_collection(str(collection_source))

    def get_database(self):
        self._enq = None
        self._parser = None
        db = False

        if self.service is not None:
            self.reload_collections()
            db = get_collated_database(self._collections)

        if db:
            self._enq = xapian.Enquire(db)
            self._parser = xapian.QueryParser()
            self._parser.set_default_op(xapian.Query.OP_AND)
            self._parser.add_prefix("lkd", "LKD")
            self._parser.set_database(db)
        return db

    def chooser_response(self, res):
        res = str(res)
        if res:
            self.reload_collections()
            i, _ = self.get_index(res)
            self.build_collection(res)
            self.collectionAdded.emit(i, res)
        return

    def reload_database(self):
        if self.db:
            self.db.close()
        self.db = self.get_database()
        self.databaseReloaded.emit()
        return

    def document_changed(self, docid):
        self.db.reopen()
        self.documentChanged.emit(docid)
        return

    def partial_query(self, qstr, n=0, sort_key=DbEnums.V_SORTING):
        q = self._parser.parse_query(
            qstr,
            self.DEFAULT_QUERY_FLAGS)

        self._enq.set_sort_by_value(sort_key, False)
        self._enq.set_query(q)
        return self._enq.get_mset(0, n or self.db.get_doccount())

    #~ def query(self, qstr, flags):
        #~ return
