import xapian
import os.path
from os.path import exists

from gi.repository import Gio

from ..collectors.cdda import CddaCollector
from ..collectors.filesystem import FileSystemCollector

from totallywired.utils.misc import (
    StdoutColors,
    ellipsis,
    parse_date_as_year_only)

from totallywired.utils.paths import (
    is_cdda_gvfs, 
    get_cdda_gvfs,
    source2collection)


class Collection(object):

    ICON_SIZE = 48

    def __init__(self, manager, source):
        self.manager = manager
        self.collector = None
        self._source = source
        self.working = False

        if self.live:
            db = self.get_readonly_db()
            if self.is_hidden(db):
                self.state = "hidden"
            else:
                self.state = "ready"
        else:
            self.state = "not-connected"
        return

    def __str__(self):
        r = "%s [%s] <%s>" % (self.source, ellipsis(self.xpath, 30), self.state)
        if self.live:
            return "%s%s%s" % (StdoutColors.OKBLUE, r, StdoutColors.ENDC)
        return "%s%s%s" % (StdoutColors.WARNING, r, StdoutColors.ENDC)

    def _get_mount(self):
        source = self.source
        for mnt in self.manager.volume_monitor.get_mounts():
            ident = mnt.get_root().get_path()
            if ident is None:
                continue
            elif source.startswith(ident):
                return mnt
        return None

    def mount(self):
        if exists(self.source):
            return True

        self.collector.read_metadata()

        for vol in self.manager.volume_monitor.get_volumes():
            if vol.get_name() == self.collector.volume_name:
                if vol.can_mount():
                    break
                else:
                    return False
        else:
            return False

        print("Attempting mount of %s..." % self.source)

        mount_operation = Gio.MountOperation()
        mount_operation.set_anonymous(True)

        def finished_cb(volume, result):
            volume.mount_finish(result)
            return

        vol.mount(
            mount_operation,
            finished_cb)
        return True

    def get_volume_name(self):
        if self.collector is None:
            return ""

        if not self.collector.volume_name:
            if self.source.startswith("/home"):
                self.collector.volume_name = "home"
            else:
                mnt = self._get_mount()
                if mnt is not None:
                    vol = mnt.get_volume()
                    if vol is not None:
                        self.collector.volume_name = vol.get_name()
        return self.collector.volume_name

    def to_dict(self):
        raise NotImplementedError("Collection needs to implement 'to_dict()'")

    def get_writable_db(self):
        try:
            db = xapian.WritableDatabase(self.xpath, xapian.DB_OPEN)
        except Exception as e:
            print(e)
            return False
        return db

    def get_readonly_db(self):
        try:
            db = xapian.Database(self.xpath)
        except Exception as e:
            print(e)
            return False
        return db

    @property
    def source(self):
        if is_cdda_gvfs(self._source):
            return self.get_cdda_gvfs()
        return self._source

    @property
    def xpath(self):
        return source2collection(self.source)

    @property
    def live(self):
        return os.path.exists(self.source)

    @property
    def hidden(self):
        return "hidden" in self.state

    @property
    def doccount(self):
        if not self.xpath or not exists(self.xpath):
            return 0
        try:
            db = xapian.Database(self.xpath)
        except xapian.DatabaseOpeningError:
            return 0
        return db.get_doccount()

    @property
    def icon_source(self):
        theme = self.manager._icon_theme
        for ico in str(self.icon).split(" "):
            icon = theme.lookup_icon(ico, self.ICON_SIZE, 0)
            if icon is not None:
                return icon.get_filename()
        return ""

    def set_hidden(self, db, hidden):
        if hidden:
            db.set_metadata("hidden", "y")
            self.state = self.state.replace("ready", "hidden")
        else:
            db.set_metadata("hidden", "n")
            self.state = self.state.replace("hidden", "ready")
        return

    def is_hidden(self, db):
        try:
            return db.get_metadata("hidden") == "y"
        except:
            pass
        return False


class StandardCollection(Collection):

    def __init__(self, manager, source):
        Collection.__init__(self, manager, source)
        self.collector = FileSystemCollector(self)
        return

    @property
    def title(self):
        source = self.source
        if source.startswith('/home'):
            try:
                user = source.split(os.sep)[2]
            except:
                return source
            return "%s's %s" % (user, os.path.split(source)[1])
        mnt = self._get_mount()
        if mnt is not None:
            folder = os.path.split(self.source)[1]
            if not folder:
                return mnt.get_volume().get_name()
            return "%s on %s" % (folder, mnt.get_volume().get_name())
        else:
            folder = os.path.split(source)[-1]
            return "%s on %s" % (folder, self.collector.volume_name)

        return self.source

    @property
    def icon(self):
        source = self.source
        if source.find('/home') > -1 and source.find('.') == -1:
            # crudely handles xdg dir icons
            # XXX: this should use some xdg accurate lib!??? needs research
            if source.endswith('Music'):
                return "folder-music"
            elif source.endswith('Desktop'):
                return "user-desktop"   # not xdg?
            elif source.endswith('Documents'):
                return "folder-documents"
            elif source.endswith('Public'):
                return "folder-publicshare"
            elif source.endswith('Downloads'):
                return "folder-download"
            elif source.endswith('Pictues'):
                return "folder-pictures"
            elif source.endswith('Templates'):
                return "folder-templates"
            elif source.endswith('Videos'):
                return "folder-videos"
            return "folder"

        mnt = self._get_mount()
        if mnt is not None:
            return mnt.get_icon().to_string().replace(". GThemedIcon ", "")
        return "folder"

    def to_dict(self):
        c = self.collector
        _dict = {
            "Source": self.source,
            "Xpath": self.xpath,
            "Title": self.title,
            "Icon": self.icon,
            "IconSource": self.icon_source,
            "State": self.state,
            "Doccount": str(self.doccount),
            "Progress": str(c.progress),}
        return _dict


class CddaCollection(Collection):

    def __init__(self, manager, source):
        Collection.__init__(self, manager, source)

        if self.live:
            db = self.get_readonly_db()
            if self.is_hidden(db):
                self.state = "hidden|cdda|no-release"
            else:
                self.state = "ready|cdda|no-release"
        else:
            self.state = "not-connected|cdda|no-release"

        self.collector = CddaCollector(self)
        self.collector.read_metadata()
        return

    @property
    def source(self):
        return get_cdda_gvfs()

    @property
    def title(self):
        if self.collector.artist and self.live:
            c = self.collector
            title = "%s by %s (%s)" % (
                c.album, c.artist, parse_date_as_year_only(c.date))
            return title
        return "Audio CD"

    @property
    def icon(self):
        mnt = self._get_mount()
        if mnt is not None:
            return mnt.get_icon().to_string().replace(". GThemedIcon ", "")
        return "folder"

    @property
    def disc_count(self):
        if self.collector.disc_count:
            return self.collector.disc_count
        return 1

    @property
    def disc_num(self):
        if self.collector.disc_num:
            return self.collector.disc_num
        return 0

    def to_dict(self):
        c = self.collector
        _dict = {
            "Source": self.source,
            "Xpath": self.xpath,
            "Title": self.title,
            "Icon": self.icon,  # icon name 
            "IconSource": self.icon_source, # themed icon path
            "State": self.state,
            "Doccount": str(self.doccount),
            "Progress": str(c.progress),
            "Album": c.album,
            "Artist": c.artist,
            "Date": c.date,
            "DiscCount": str(self.disc_count),
            "DiscNum": str(self.disc_num),
            "AlbumArt": c.album_art,
            "Tags": c.tags}
        return _dict


#~ class RadioStationCollection(Collection):
#~ 
    #~ def __init__(self, manager, source):
        #~ Collection.__init__(self, manager, source)
        #~ self.collector = RadioStationCollector(self)
        #~ self.collector.read_metadata()
        #~ return
