import os
import string
import xapian
import threading

from gi.repository import GLib

from .enums import DbEnums
from .filenames import FilenameCollector

import totallywired.utils.lastfm as lastfm

from totallywired.utils.db import (
    get_term_generator,
    get_qterm,
    get_artist_qterm,
    get_album_qterm,
    purge_all_terms)

from totallywired.utils.misc import (
    get_nice_hms,
    remove_diacritical_marks,
    parse_date_as_year_only)


class CddaCollector(FilenameCollector):

    def __init__(self, collection):
        FilenameCollector.__init__(self, collection)
        self.reset()
        return

    def reset(self):
        FilenameCollector.reset(self)
        self.progress = -1
        self.mbid = ""
        self.release = {}
        self.query = ""
        self.artist = ""
        self.album = ""
        self.date = ""
        self.disc_count = 1
        self.disc_num = 0
        self.tags = ""
        self.album_art = ""
        return

    def save_metadata(self, db):
        FilenameCollector.save_metadata(self, db)
        db.set_metadata("artist", self.artist)
        db.set_metadata("album", self.album)
        db.set_metadata("date", self.date)
        db.set_metadata("discNum", str(self.disc_num))
        db.set_metadata("discCount", str(self.disc_count))
        db.set_metadata("albumArt", self.album_art)
        db.set_metadata("tags", self.tags)
        return

    def read_metadata(self):
        db = FilenameCollector.read_metadata(self)
        if not db:
            return False

        self.artist = str(db.get_metadata("artist"))
        self.album = str(db.get_metadata("album"))
        self.date = str(db.get_metadata("date"))
        self.disc_count = str(db.get_metadata("discCount"))
        self.disc_num = str(db.get_metadata("discNum"))
        self.album_art = str(db.get_metadata("albumArt"))
        self.tags = str(db.get_metadata("tags")).decode("utf-8")
        return True

    def _setup_with_discid(self):
        did = lastfm.get_cdda_discid()
        if not did:
            return False
        mbid = lastfm.get_mbid_from_discid(did)
        if not mbid: 
            return False

        try:
            release = lastfm.mb.get_release_by_id(
                mbid,
                includes=["artists", "artist-credits", "recordings"])
        except Exception as e:
            print('Could not find release!')
            print(e)
            return False

        self.release = release["release"]
        return mbid

    def _setup_with_mbquery(self, query):
        try:
            if hasattr(lastfm.mb, "release_search"):
                res = lastfm.mb.release_search(query=query, limit=1)
            elif hasattr(lastfm.mb, "search_releases"):
                res = lastfm.mb.search_releases(query=query, limit=1)
        except Exception as e:
            print("Could not lookup release")
            print(e)
            return False

        if res['release-list']:
            mbid = res['release-list'][0]['id']
            release = lastfm.mb.get_release_by_id(
                mbid,
                includes=["artists", "artist-credits", "recordings"])
            self.release = release['release']
            return mbid

        return False

    def _get_artist_tags(self):
        # get artist, for tags only atm
        artist_mbid = self.release["artist-credit"][0]["artist"]["id"]
        try:
            artist = lastfm.mb.get_artist_by_id(artist_mbid, includes=["tags",])
        except Exception as e:
            print("Error encountered looking up artist (%s)" % e)
            return self._setup_failed()

        artist = artist["artist"]

        if "tag-list" in artist:
            tags = []
            for tag in artist["tag-list"]:
                tags.append(tag["name"].replace("/", ", "))

            return string.join(tags, ", ").lower()
        return ""

    def setup(self, query, disc_num):
        self.disc_num = disc_num

        if not query and self.release:
            print("A release exists, recycling attrs (this should occur when user changes disc_num)")
            print("Using mbid:", self.mbid, self.disc_num)
            # no query but release info exists, we can be lazy and use
            # current attr state
            return self.mbid
        elif not query and not self.release:
            # check for existing metadata
            if self.artist or self.album:
                query = '"%s" AND artist:"%s"' % (self.album, self.artist)
                print("Forming query from saved metadata...")
                print(query)
                print("Attempting said query...")
                self.mbid = self._setup_with_mbquery(query)
            else:   # try lookup based on discid
                print("Attempting lookup from discid...")
                self.mbid = self._setup_with_discid()
        elif query:
            print("Attempting MusicBrainz query:", query)
            self.mbid = self._setup_with_mbquery(query)
        else:
            print("eeeeek!")
            print(self.query)
            print(self.release)
            return False

        if not self.mbid:
            return self._setup_failed()

        self.progress = 0.5
        self.send_message({"progress-update": True})

        try:
            self.tags = self._get_artist_tags()
        except Exception as e:
            print e
            self.tags = ""

        self.progress = 0.7
        self.send_message({"progress-update": True})

        try:
            self.album_art = lastfm.get_artwork(self.mbid)
        except Exception as e:
            print e
            self.album_art = ""

        self.progress = 0.9
        self.send_message({"progress-update": True})

        print()
        return True

    def _setup_failed(self):
        self.progress = 0.333
        self.send_message({"progress-update": True})

        L = FilenameCollector.run(
            self,
            terms=["__cdda",],
            progress_offset=self.progress,
            progress_extent=1.0)

        self.progress = 1.0
        self.send_message({"progress-update": True})

        c = self.collection
        c.state = "ready|cdda|no-release"
        c.dococunt = L
        self.send_finished_message()
        return False

    @staticmethod
    def index_document(db, term_generator, f, artist, album, date, track, tags):
        uri = GLib.filename_to_uri(f, None)
        pl = db.postlist(get_qterm(uri))
        plitem = None
        try:
            plitem = next(pl)
        except StopIteration:
            print(("No doc with qterm: %s. Will now create..." % get_qterm(uri)))

        if plitem is not None:
            doc = db.get_document(plitem.docid)
            purge_all_terms(doc)
        else:
            doc = xapian.Document()

        doc.set_data(uri)
        doc.add_term(get_qterm(uri), 0)
        # special term for easy cdda lookup/queries
        doc.add_term("__cdda", 0)

        pos = track['position']
        doc.add_value(DbEnums.V_TRACKNUM, pos)
        doc.add_value(DbEnums.V_DATE, date)

        recording = track["recording"]

        doc.add_value(
            DbEnums.V_DURATION_FSTR,
            get_nice_hms(int(recording['length']) / 1000))

        term_generator.set_document(doc)

        recording_artist = str(recording['artist-credit-phrase'])
        doc.add_value(DbEnums.V_ARTIST, recording_artist)
        term_generator.index_text_without_positions(recording_artist, DbEnums.W_ARTIST)
        doc.add_term(get_artist_qterm(remove_diacritical_marks(recording_artist)))

        album_artist = str(artist)
        doc.add_value(DbEnums.V_AARTIST, album_artist)
        term_generator.index_text_without_positions(album_artist, DbEnums.W_AARTIST)
        doc.add_term(get_artist_qterm(remove_diacritical_marks(album_artist)))

        doc.add_value(DbEnums.V_ALBUM, album)
        term_generator.index_text_without_positions(album, DbEnums.W_ALBUM)
        doc.add_term(get_album_qterm(remove_diacritical_marks(album)))

        title = unicode(recording['title'])
        doc.add_value(DbEnums.V_TITLE, title)
        term_generator.index_text_without_positions(title, DbEnums.W_TITLE)

        # tags
        if tags:
            doc.add_value(DbEnums.V_TAGS, tags)
            # add searchable terms for each tag
            for tag in tags.split(","):
                # check if we can convert, say, u"Fanto^mas" to u"Fantomas"
                # for easy English keyboard searching... store both terms
                ndmt = remove_diacritical_marks(tag.strip())
                term_generator.index_text(ndmt, DbEnums.W_TAGS)

        sort_v = (
            recording_artist + album +
            xapian.sortable_serialise(float(pos)).decode('latin1') +
            title)
        doc.add_value(DbEnums.V_SORTING, sort_v)

        if plitem is not None:
            db.replace_document(plitem.docid, doc)
        else:
            db.add_document(doc)
        return

    def index_all(self):
        release = self.release

        if not release:
            return

        source = self.source

        if not os.path.exists(self.xpath):
            print("The xpath does not exist, creating new cdda database at:", self.xpath)
            db = xapian.WritableDatabase(self.xpath, xapian.DB_CREATE)
        else:
            db = xapian.WritableDatabase(self.xpath, xapian.DB_OPEN)

        term_generator = get_term_generator(db)

        # check the disc_num is within reason
        self.disc_num = min(len(release['medium-list'])-1, self.disc_num)

        cd_tracks = os.listdir(source)
        print cd_tracks
        mb_tracks = release['medium-list'][self.disc_num]['track-list']

        self.disc_count = len(release["medium-list"])
        self.artist = str(release["artist-credit-phrase"])
        self.album = str(release["title"])
        self.date = parse_date_as_year_only(str(release['date']))

        print()
        print("Updating database with collected metadata...")

        for i, track in enumerate(mb_tracks):
            try:
                t_path = os.path.join(source, cd_tracks[i])
            except IndexError as e:
                print("MB track list did not match real track list")
                print(e)
                break
            CddaCollector.index_document(
                db, term_generator, t_path,
                self.artist, self.album, self.date, track, self.tags)

        self.save_metadata(db)

        db.commit()
        db.close()
        print("Database update complete")
        print()
        return len(cd_tracks)

    def run(self, query="", disc_num=0):
        self.progress = 0.0
        self.send_message({"progress-update": True})

        def run_async(query, disc_num):
            res = self.setup(query, disc_num)
            if not res:
                return

            self.index_all()

            self.progress = 1
            self.send_message({"progress-update": True})

            self.teardown()

            c = self.collection
            c.state = "ready|cdda|release-found"
            self.send_finished_message({})
            return

        job = threading.Thread(
            group=None,
            target=run_async,
            name="CddaCollector-%s" % (threading.active_count()+1),
            args=(query, disc_num),
            kwargs={},
            verbose=None)
        job.daemon = True
        job.start()
        return

    def run_with_metadata(self, metadata):
        self.release = metadata
        self.index_all()
        c = self.collection
        c.state = "ready|cdda|release-found"
        self.send_finished_message({})
        return

    def teardown(self):
        self.progress = -1
        return
