import os
import xapian
import wave
import contextlib

import gio
from gi.repository import GLib

from .enums import DbEnums

from totallywired.utils.misc import get_nice_hms
from totallywired.utils.paths import recursive_find_by_ext

from totallywired.utils.db import (
    get_qterm,
    get_artist_qterm,
    get_album_qterm,
    get_term_generator)


class FilenameCollector(object):

    UNKNOWN = 'Unknown'

    def __init__(self, collection):
        self.collection = collection
        self.volume_name = ""
        self.progress = -1
        return

    def reset(self):
        print("Resetting CddaCollector attrs...")
        return

    @staticmethod
    def try_get_wave_length(path):
        try:
            with contextlib.closing(wave.open(path,'r')) as f:
                frames=f.getnframes()
                rate=f.getframerate()
                duration=frames/float(rate)
        except:
            return 0
        return duration

    @property
    def source(self):
        return self.collection.source

    @property
    def xpath(self):
        return self.collection.xpath

    def send_message(self, msg={}):
        self.collection.manager.on_collector_message(
            self, msg)
        return

    def send_finished_message(self, msg={}):
        self.progress = -1
        self.collection.manager.on_collector_finished(
            self, msg)
        return

    def setup(self):
        self.audiofiles = recursive_find_by_ext(self.source)
        working_path = self.xpath.rstrip('/') + ".indexing"
        db = xapian.WritableDatabase(
            working_path, xapian.DB_CREATE_OR_OVERWRITE)
        self.term_generator = get_term_generator(db)
        self.db = db
        self.working_path = working_path
        return

    @staticmethod
    def index_document(doc, term_generator, f, terms=[]):
        # index a file that has no metadata
        uri = GLib.filename_to_uri(f, None)

        doc.set_data(uri)
        doc.add_term(get_qterm(uri), 0)
        doc.add_term(get_artist_qterm(FilenameCollector.UNKNOWN), 0)
        doc.add_term(get_album_qterm(FilenameCollector.UNKNOWN), 0)
        term_generator.set_document(doc)
        term_generator.index_text_without_positions(os.path.split(f)[1], 2)

        for term in terms:
            doc.add_term(str(term), 5)

        t = os.path.split(f)[-1]
        doc.add_value(DbEnums.V_TITLE, t)
        doc.add_value(DbEnums.V_ARTIST, FilenameCollector.UNKNOWN)
        doc.add_value(DbEnums.V_AARTIST, FilenameCollector.UNKNOWN)
        doc.add_value(DbEnums.V_ALBUM, FilenameCollector.UNKNOWN)
        doc.add_value(DbEnums.V_SORTING, FilenameCollector.UNKNOWN)

        dur = FilenameCollector.try_get_wave_length(f)
        if dur:
            # index display duration
            doc.add_value(
                DbEnums.V_DURATION_FSTR,
                get_nice_hms(dur))
        return

    def index_all(self, terms, progress_start, progress_end):
        db = self.db
        term_generator = self.term_generator
        L = len(self.audiofiles)
        pr = progress_end - progress_start  # the progress 'range'

        for i, f in enumerate(self.audiofiles):
            doc = xapian.Document()
            FilenameCollector.index_document(
                doc,
                term_generator,
                f, terms)
            db.add_document(doc)

            self.progress = progress_start + (pr * i / L)
            self.send_message({"progress-update":True,})

        self.save_metadata(db)

        self.db.commit()
        self.db.close()
        return L

    def teardown(self):
        # FIXME: all indexers should do this.  move shutil code into 
        #        BaseIndexer class.
        import shutil
        if os.path.exists(self.xpath):
            # rm old db tree
            shutil.rmtree(self.xpath)
        # mv new db to old db location
        shutil.move(self.working_path, self.xpath)
        # return to idle state
        self.progress = -1
        return

    def read_metadata(self):
        #~ print(self.xpath)
        if not self.xpath or not os.path.exists(self.xpath):
            #~ print("No database currently exists at:", self.xpath)
            #~ print("Skipping metadata read...")
            return False

        try:
            db = xapian.Database(self.xpath)
        except xapian.DatabaseOpeningError:
            return False
        self.volume_name = db.get_metadata("volName")
        return db

    def save_metadata(self, db):
        #~ print("Saving metadata to collection db...")
        db.set_metadata("volName", self.collection.get_volume_name())
        return

    def run(self, terms=[], progress_offset=0.0, progress_extent=1.0):
        FilenameCollector.setup(self)
        FilenameCollector.index_all(
            self, terms,
            progress_offset,
            progress_extent)
        FilenameCollector.teardown(self)

        c = self.collection
        c.state = "ready"
        self.send_finished_message({})
        return
