import time
import string
import xapian
import os.path
import traceback
import threading

from gi.repository import GLib

from .enums import (
    DbEnums,
    term2value,
    term2weight)

from totallywired.utils.db import (
    get_qterm,
    get_artist_qterm,
    get_album_qterm,
    get_term_generator)

from totallywired.utils.misc import (
    get_nice_hms,
    remove_diacritical_marks,
    parse_date_as_year_only)

from totallywired.utils.caching import Caches
from totallywired.utils.paths import recursive_find_by_ext

from . import get_metadata
from filenames import FilenameCollector


def parse_attr(attr):
    if isinstance(attr, (list, tuple)):
        return unicode(attr[0])
    return unicode(attr)

class FileSystemCollector(FilenameCollector):

    UNKNOWN = "Unknown"

    def __init__(self, collection):
        FilenameCollector.__init__(self, collection)
        self.liked = []
        self.disliked = []
        return

    def _remember_lkd_tracks(self, liked, disliked):
        if not os.path.exists(self.xpath):
            return
        db = xapian.Database(self.xpath)
        pl = db.postlist(DbEnums.T_LIKE)
        while 1:
            try:
                plitem = next(pl)
                liked.append(db.get_document(plitem.docid).get_data())
            except StopIteration:
                break

        pl = db.postlist(DbEnums.T_DISLIKE)
        while 1:
            try:
                plitem = next(pl)
                disliked.append(db.get_document(plitem.docid).get_data())
            except StopIteration:
                break

        db.close()
        print('LKD term count: %s' % (len(liked) + len(disliked)))
        return

    def setup(self):
        self._remember_lkd_tracks(self.liked, self.disliked)

        self.audiofiles = recursive_find_by_ext(self.source)
        working_path = self.xpath.rstrip('/') + ".indexing"
        db = xapian.WritableDatabase(
            working_path, xapian.DB_CREATE_OR_OVERWRITE)
        self.tg = get_term_generator(db)
        self.db = db
        self.working_path = working_path
        return

    @staticmethod
    def index_tracknum(doc, tg, kmap, data, k=DbEnums.T_TRACKNUM):
        # translate the db key to an "encoding specific key" (esk)
        esk = kmap[k]
        if esk in data:
            t = parse_attr(data[esk])
        else:
            t = ""

        if t:
            if not t.isnumeric() and t.find('/') > 0:
                t = t.split('/')[0]
                t = unicode(t.lstrip('0'))
            elif t.startswith('('):   # m4a (tack number,track total)
                t = t.strip('()')
                t = t.split(",")[0]

            doc.add_value(term2value[k], t)
            t = xapian.sortable_serialise(float(t)).decode('latin1')
        return t

    @staticmethod
    def index_year(doc, tg, kmap, data, k=DbEnums.T_DATE):
        # translate the db key to an "encoding specific key" (esk)
        esk = kmap[k]
        t_found = True
        if esk in data:
            t = parse_attr(data[esk])
        else:
            t = ""
            t_found = False

        if t_found:
            doc.add_value(term2value[k], parse_date_as_year_only(t))
        return t

    @staticmethod
    def index_str(doc, tg, kmap, key, data, value=False, fallback=""):
        # translate the db key to an "encoding specific key" (esk)
        esk = kmap[key]
        t_found = True
        if esk in data:
            t = parse_attr(data[esk])
        else:
            t = fallback
            t_found = False

        doc.add_value(value or term2value[key], t)

        if t_found:
            # check if we can convert, say, u"Fanto^mas" to u"Fantomas"
            # for easy English keyboard searching... store both terms
            ndmt = remove_diacritical_marks(t)
            tg.index_text_without_positions(ndmt, term2weight[key])
            return ndmt
        return t

    @staticmethod
    def index_tags(doc, tg, kmap, k, data):
        # translate the db key to an "encoding specific key" (esk)
        esk = kmap[k]
        t_found = True
        if esk in data:
            tags = parse_attr(data[esk])
        else:
            tags = ""
            t_found = False

        if t_found:
            w = term2weight[k]
            tags = tags.lower().replace("/", ", ")
            doc.add_value(term2value[k], tags)
            # add searchable terms for each tag
            for tag in tags.split(","):
                # check if we can convert, say, u"Fanto^mas" to u"Fantomas"
                # for easy English keyboard searching... store both terms
                ndmt = remove_diacritical_marks(tag.strip())
                tg.index_text(ndmt, w)
            return tags
        return []

    @staticmethod
    def index_document(doc, tg, f, data, kmap, liked=[], disliked=[]):
        if data is None:
            # if no metadata is available, use simple
            # fallback indexer
            FilenameCollector.index_document(doc, tg, f)
            return

        uri = GLib.filename_to_uri(f, None)

        if uri in liked:
            doc.add_value(DbEnums.V_LIKE, "1")
            doc.add_term(DbEnums.T_LIKE)
        elif uri in disliked:
            doc.add_value(DbEnums.V_LIKE, "2")
            doc.add_term(DbEnums.T_DISLIKE)

        # intex uri qterm
        doc.set_data(uri)
        doc.add_term(get_qterm(uri), 0)

        # index display duration
        doc.add_value(
            DbEnums.V_DURATION_FSTR,
            get_nice_hms(data.info.length))

        tg.set_document(doc)
        # tags / genre
        FileSystemCollector.index_tags(
            doc, tg, kmap, DbEnums.T_TAGS, data)

        sorter = []

        # index artist name
        artist = FileSystemCollector.index_str(
            doc, tg, kmap, DbEnums.T_ARTIST, data)
        doc.add_term(get_artist_qterm(artist))

        # index album-artist name
        aartist = FileSystemCollector.index_str(
            doc, tg, kmap, DbEnums.T_AARTIST, data)
        if aartist:
            doc.add_term(get_artist_qterm(aartist))
            sorter.append(aartist)
        else:
            FileSystemCollector.index_str(
                doc, tg, kmap, DbEnums.T_ARTIST, data, value=DbEnums.V_AARTIST)
            sorter.append(artist)

        # index album name
        sorter.append(FileSystemCollector.index_str(
            doc, tg, kmap, DbEnums.T_ALBUM, data))
        doc.add_term(get_album_qterm(sorter[-1]))

        # index track number
        sorter.append(FileSystemCollector.index_tracknum(
            doc, tg, kmap, data))

        # index track name
        sorter.append(FileSystemCollector.index_str(
            doc, tg, kmap, DbEnums.T_TITLE, data,
            fallback=os.path.split(f)[-1]))

        # index date
        sorter.append(FileSystemCollector.index_year(
            doc, tg, kmap, data))

        # add sorter
        doc.add_value(DbEnums.V_SORTING, string.join(sorter).replace(" ", ""))
        return

    @staticmethod
    def index_document_easy(db, tg, f, liked=[], disliked=[]):
        data, kmap = get_metadata(f)
        doc = xapian.Document()
        FileSystemCollector.index_document(
            doc, tg,
            f,
            data,
            kmap,
            liked,
            disliked)
        db.add_document(doc)
        return

    def index_all(self):
        self.progress = 0
        self.send_message({"progress-update":True,})

        db = self.db
        liked = self.liked
        disliked = self.disliked
        t0 = time.time()
        L = float(len(self.audiofiles))
        tg = self.tg

        for i, f in enumerate(self.audiofiles):
            doc = xapian.Document()
            try:
                data, kmap = get_metadata(f)
                FileSystemCollector.index_document(
                    doc, tg,
                    f,
                    data,
                    kmap,
                    liked,
                    disliked)
            except:
                tbpath = os.path.join(Caches.ERRORS, "tw.filesystem.collector.traceback.%s" % time.time())
                with open(tbpath, "w") as f:
                    print("Exception in user code.  See '%s' for traceback." % tbpath)
                    traceback.print_exc(file=f)
                continue

            db.add_document(doc)

            tc = time.time()
            if tc - t0 >= 2:
                t0 = tc
                self.progress = i / L
                self.send_message({"progress-update":True,})

        self.save_metadata(db)

        self.db.commit()
        self.db.close()
        self.progress = 1
        self.send_message({"progress-update":True,})
        return L

    def run(self):
        def run_async():
            self.setup()
            self.index_all()
            self.teardown()

            c = self.collection
            c.state = "ready"
            self.send_finished_message({})
            return

        job = threading.Thread(
            group=None,
            target=run_async,
            name="FileSystemCollector-%s" % (threading.active_count()+1),
            args=(),
            kwargs={},
            verbose=None)
        job.daemon = True
        job.start()
        return
