import os
import pyinotify
import xapian

from gi.repository import Gio
from gi.repository import GLib

from totallywired.db.collectors.filesystem import FileSystemCollector
from totallywired.utils.misc import ellipsis

from totallywired.utils.db import (
    get_collated_database,
    get_live_sources,
    difflasttime,
    get_writable_twxdbs_for_uris,
    get_qterm,
    mklasttime,
    qterm_in_db,
    get_term_generator,
    get_twxdb_for_uri)

from totallywired.utils.paths import (
    source2collection,
    is_cdda_path, 
    is_trash_path,
    recursive_find_by_ext)


def get_watchable_sources():
    return [s for s in get_live_sources() if not is_cdda_path(s)]


class EventHandler(pyinotify.ProcessEvent):

    def __init__(self, f_change_cb, d_change_cb):
        self.f_change_cb = f_change_cb
        self.d_change_cb = d_change_cb
        return

    def handle_update(self, event):
        if event.dir:
            self.d_change_cb(event)
        else:
            self.f_change_cb(event)

    def process_IN_ATTRIB(self, event):
        print('ATTRIB', event.pathname)
        self.handle_update(event)

    def process_IN_MOVED_FROM(self, event):
        print('MV_FROM', event.pathname)
        self.handle_update(event)

    def process_IN_MOVED_TO(self, event):
        print('MV_TO', event.pathname)
        self.handle_update(event)

    #~ def process_IN_CREATE(self, event):
        #~ print 'OPEN', event.pathname, event.name

    def process_IN_DELETE(self, event):
        print('DEL', event.pathname)
        self.handle_update(event)

    #~ def process_IN_MODIFY(self, event):
        #~ print 'OPEN', event.pathname, event.name



class SourceMonitor(object):

    DIRWATCH_EVENT_MASK = (#pyinotify.IN_CREATE |
                           pyinotify.IN_DELETE |
                           #pyinotify.IN_MODIFY |
                           pyinotify.IN_ATTRIB |
                           pyinotify.IN_MOVED_TO |
                           pyinotify.IN_MOVED_FROM)

    DIRWATCH_TIMEOUT = 3000

    MOUNT_ADDED_EVENT = "added"
    MOUNT_REMOVED_EVENT = "removed"

    def __init__(self):
        self.is_running = True

        # volume manager
        self._prev_e_tag = ""
        self.volume_monitor = Gio.VolumeMonitor()
        self.volume_monitor.connect("mount-added", self._mount_changed, self.MOUNT_ADDED_EVENT)
        self.volume_monitor.connect("mount-removed", self._mount_changed, self.MOUNT_REMOVED_EVENT)

        # pyionotify watcher
        self._watchlist = {}
        self._pjob = False
        self._fevents = set()
        self._needs_snapshots = False

        self.watchmanager = pyinotify.WatchManager()
        eh = EventHandler(self._handle_file_change, self._handle_dir_change)
        self.notifier = pyinotify.Notifier(self.watchmanager, eh)

        # cdda audio info
        self.cdda_state = "not-connected"
        self.cdda_album_title = "Audio CD"

        # job queue
        self._jqueue = []

        GLib.timeout_add(1000, self._init)
        return

    def _init(self):
        print("Starting dir watch...")
        self._init_dirwatch(self.notifier, self.watchmanager)
        print("Diff-ing filesystems...")
        self.diff_filesystems()
        return

    def _handle_file_change(self, event):
        # check if we are dealing with a movement into a .Trash folder
        if is_trash_path(event.pathname):
            return
        self._fevents.add(GLib.filename_to_uri(event.pathname, None))
        return

    def _handle_dir_change(self, event):
        pn = event.pathname
        # check if we are dealing with a movement into a .Trash folder
        if is_trash_path(pn):
            return
        if os.path.exists(pn):
            files = (GLib.filename_to_uri(f, None) for f in recursive_find_by_ext(pn))
            self._fevents |= set(files)
        else:
            uri = GLib.filename_to_uri(pn, None)
            db, _ = get_twxdb_for_uri(uri)
            e = xapian.Enquire(db)
            e.set_query(xapian.Query(""))
            mset = e.get_mset(0, db.get_doccount())
            uris = (u for u in (db.get_document(m.docid).get_data() for m in mset) if u.startswith(uri))
            self._fevents |= set(uris)
            print("Number of events: %s" % len(self._fevents))
            db.close()
        return

    def _process_batch(self):
        if not len(self._fevents):
            return

        # preprocess to determine reqd dbs for the current batch of
        # files to process
        dbs = get_writable_twxdbs_for_uris(self._fevents)

        # for the jobs relating to each db, either add or remove docs\
        for src, (db, uris) in dbs.items():
            tg = get_term_generator(db)
            for uri in uris:
                f = GLib.filename_from_uri(uri, "")
                qt = get_qterm(uri)
                if os.path.exists(f):
                    if not qterm_in_db(db, qt):
                        FileSystemCollector.index_document_easy(db, tg, f)
                        print("Creating Document: %s", ellipsis(f, 30))
                    else:
                        print('"%s" is already indexed... Skipping' % ellipsis(f, 30))
                else:
                    if qterm_in_db(db, qt):
                        db.delete_document(qt)
                        print("Removing Document: %s" % ellipsis(f, 30))
                    else:
                        print('"%s" is not indexed... Skipping' % ellipsis(f, 30))

            print('NewDoccount: %s (%s)' % (db.get_doccount(), ellipsis(src, 30)))
            db.commit()
            db.close()

        self._fevents.clear()
        # fixme, can we do something finer grained here??
        self._needs_snapshots = True
        self.do_database_reload()
        return True

    def _init_dirwatch(self, notifier, watchmanager):
        for src in get_watchable_sources():
            self._add_watch(watchmanager, src)

        GLib.timeout_add(
            self.DIRWATCH_TIMEOUT,
            self._process_io_events,
            notifier)
        return

    def _update_dirwatch(self, src, event):
        if event == self.MOUNT_REMOVED_EVENT and src in self._watchlist:
            del self._watchlist[src]
        elif event == self.MOUNT_ADDED_EVENT:
            self._add_watch(self.watchmanager, src)
            diff, _ = difflasttime(source2collection(src))
            self._fevents |= set([GLib.filename_to_uri(f, None) for f in diff])
        return

    def _add_watch(self, watchmanager, src):
        wdd = watchmanager.add_watch(
            src,
            self.DIRWATCH_EVENT_MASK,
            rec=True,
            auto_add=True)
        self._watchlist[src] = wdd
        return

    def _process_io_events(self, notifier):
        try:
            notifier.process_events()
            if notifier.check_events(200):
                notifier.read_events()
        except KeyboardInterrupt:
            print("\nKeyboardInterrupt!")
            self.quit()
            return False

        if self._pjob:
            self._process_batch()
        self._pjob = not self._pjob
        return True

    def _get_volume_identifier(self, drive):
        label = drive.get_identifier(Gio.VOLUME_IDENTIFIER_KIND_LABEL)
        uuid = drive.get_identifier(Gio.VOLUME_IDENTIFIER_KIND_UUID)
        return os.path.join("/media", label or uuid)

    def _is_duplicate_event(self, src, event):
        e_tag = "%s:%s" % (src, event)
        if e_tag == self._prev_e_tag:
            return True, e_tag
        return False, e_tag

    def _mount_changed(self, monitor, mount, event):
        media = mount.get_root().get_path()
        is_dup, e_tag = self._is_duplicate_event(media, event)
        if is_dup:
            print("Source 'mount' event was duplicated.  Skipping...")
            print()
            return
        self._prev_e_tag = e_tag
        self.do_mount_changed(media, event)
        return

    def diff_filesystems(self):
        all_changes = []
        for src in get_watchable_sources():
            index = source2collection(src)
            diff, curfs = difflasttime(index)
            if len(diff) > 0:
                all_changes.extend(diff)
                mklasttime(index, cur=curfs)

        if all_changes:
            print("Content on the filesystem has changed since last operation.  N of detected differences: %s" % len(all_changes))
            self._fevents |= set([GLib.filename_to_uri(f, None) for f in all_changes])
        return False

    def snapshot_filesystems(self):
        for src in get_watchable_sources():
            mklasttime(source2collection(src))
        return

    def quit(self):
        print()
        self.is_running = False
        if self._needs_snapshots:
            print('Snapshotting filesystems...')
            self.snapshot_filesystems()

        self.notifier.stop()
        return

    def do_mount_changed(self, source, event):
        return

    def do_database_reload(self):
        return
