import xapian

import os
from os.path import exists

import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from gi.repository import GLib, Gio, Gtk, GdkX11

from .monitor import SourceMonitor
from .enums import TWXDB_BASEPATH

from totallywired.utils.db import get_all_sources
from totallywired.utils.misc import ellipsis

from totallywired.db.collections.service import (
    StandardCollection,
    CddaCollection)

from totallywired.utils.paths import (
    is_cdda_path,
    is_cdda_gvfs,
    source2collection)



class StandardXPaths(object):
    CDDA = os.path.join(TWXDB_BASEPATH, "cdda")
    MUSIC = source2collection(os.path.expanduser("~%sMusic" % os.sep))


def init_standard_collections():
    # create a cdda dir
    if not os.path.exists(StandardXPaths.CDDA):
        os.mkdir(StandardXPaths.CDDA)
        # special case: we needn't create a db as these are created 
        # dynamically upon disc discovery
    print('Cdda xpath:', StandardXPaths.CDDA)

    # home 'Music' folder
    if not os.path.exists(StandardXPaths.MUSIC):
        xapian.WritableDatabase(
            StandardXPaths.MUSIC, xapian.DB_CREATE)

    print('Music xpath:', StandardXPaths.MUSIC)
    return

init_standard_collections()


class CollectionManagerService(dbus.service.Object, SourceMonitor):

    def __init__(self):
        SourceMonitor.__init__(self)
        bus_name = dbus.service.BusName('org.tw.XCollectionManagerService', bus=dbus.SessionBus())
        dbus.service.Object.__init__(self, bus_name, '/XCollectionManagerService')
        self.is_running = True
        self._icon_theme = Gtk.IconTheme.get_default()
        self._collections = []
        GLib.idle_add(self.reload_all_collections)
        print("CollectionManagerService is now running (pid:%s)" % os.getpid())
        return

    def reload_all_collections(self):
        print("\nAll Collections:")
        self._collections = []
        for c in get_all_sources():
            if c == "/cdda":
                collection = CddaCollection(self, c)
            else:
                collection = StandardCollection(self, c)
            print(collection)
            self._collections.append(collection)

        # XXX: no one is listening to this sig at the mo
        #~ self.ReloadCollections() 
        return

    def on_collector_message(self, collector, message):
        if "progress-update" in message:
            print("Collector working (%s): %.2f%%" % (collector.collection.source, round(collector.progress * 100, 2)))
            self.CollectionChanged(collector.collection.to_dict())
        return

    def on_collector_finished(self, collector, message):
        print("Collector finished: %s", ellipsis(collector.collection.title, 30))
        self.snapshot_filesystems()
        self.CollectionChanged(collector.collection.to_dict())
        self.ReloadDatabase()
        return

    def do_mount_changed(self, source, event):
        print("Mount changed: %s" % source, event)
        c = self.get_collection(source)
        if not c: return

        is_added = event == "added"
        xpath = c.xpath
        print("\nXpath: %s" % ellipsis(xpath, 60))

        c.collector.reset()
        c.collector.read_metadata()

        # handle cdda peculiarities
        if is_added and is_cdda_path(source):
            print("Looks like a new cdda has been inserted...")
            if xpath and not exists(xpath):
                c.collector.run()

        if is_added:
            c.state = c.state.replace("not-connected", "ready")
        else:
            c.state = c.state.replace("ready", "not-connected")

        print("Mount change processed: %s [%s] <%s>\n" % (source, event, c.state))

        self.CollectionChanged(c.to_dict())
        self.ReloadDatabase()
        return

    def do_database_reload(self):
        # XXX: called by monitor.py on batch updates and other fs events...
        print("Database Reload requested\n")
        self.ReloadDatabase()

    def get_collection(self, source):
        for c in self._collections:
            if source is None:
                continue
            if c.source.startswith(source):
                return c
        return False

    @dbus.service.method('org.tw.XCollectionManager', in_signature ='', out_signature='')
    def auto_mount_collections(self):
        for c in self._collections:
            if not c.live: c.mount()
        return

    @dbus.service.method('org.tw.XCollectionManager', in_signature ='si', out_signature='')
    def rerun_cdda_collector_with_mbquery(self, query, disc_num):
        for c in self._collections:
            if is_cdda_path(c.source):
                break
        else:
            return
        c.collector.run(query=query, disc_num=disc_num)
        return

    @dbus.service.method('org.tw.XCollectionManager', in_signature ='i', out_signature='')
    def rerun_cdda_collector(self, disc_num):
        for c in self._collections:
            if is_cdda_path(c.source):
                break
        else:
            return
        c.collector.run(disc_num=disc_num)
        return

    @dbus.service.signal('org.tw.XCollectionManagerService')
    def CollectionChanged(self, collection_dict):
        pass

    @dbus.service.signal('org.tw.XCollectionManagerService')
    def ReloadDatabase(self):
        pass

    @dbus.service.signal('org.tw.XCollectionManagerService')
    def ReloadCollections(self):
        pass

    @dbus.service.signal('org.tw.XCollectionManagerService')
    def ChooserResponse(self, res):
        pass

    @dbus.service.method('org.tw.XCollectionManager', in_signature ='b', out_signature='aa{ss}')
    def get_collections(self, auto_mount):
        collections = list([c.to_dict() for c in self._collections])
        if auto_mount:
            GLib.timeout_add(500, self.auto_mount_collections)
        return collections

    @dbus.service.method('org.tw.XCollectionManager', in_signature ='s', out_signature='')
    def build_collection(self, source):
        def build(collection):
            collection.collector.run()
            return

        collection = self.get_collection(str(source))
        if collection:
            GLib.timeout_add(200, build, collection)
        return

    @dbus.service.method('org.tw.XCollectionManager', in_signature='s', out_signature='')
    def toggle_hidden(self, source):
        c = self.get_collection(str(source))
        if not c: return

        db = c.get_writable_db()
        hidden = not c.is_hidden(db)
        c.set_hidden(db, hidden)
        db.close()

        self.CollectionChanged(c.to_dict())
        self.ReloadDatabase()
        return

    @dbus.service.method('org.tw.XCollectionManager', in_signature='s', out_signature='')
    def trash_collection(self, source):
        def do_trash():
            xpath = source2collection(str(source))
            print("Sending to trash: %s", xpath)
            f = Gio.file_new_for_path(xpath)
            f.trash(None)

            self.reload_all_collections()
            self.ReloadDatabase()
            return

        GLib.timeout_add(750, do_trash)
        return

    @dbus.service.method('org.tw.XCollectionManager', in_signature='', out_signature='')
    def quit(self):
        self.is_running = False
        SourceMonitor.quit(self)
        print("\nQuitting...\n")
        return

    def _display_dialog(self, wid, mode):

        def on_realize(d, parent):
            dwin = d.get_window()
            dwin.set_transient_for(parent)
            dwin.set_modal_hint(True)
            return

        d = Gtk.FileChooserDialog(
            title="Select a folder",
            parent=None,
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            buttons=(Gtk.STOCK_CANCEL,
                     Gtk.ResponseType.REJECT,
                     Gtk.STOCK_OK,
                     Gtk.ResponseType.OK))

        display = GdkX11.X11Display.get_default()
        parent = GdkX11.X11Window.foreign_new_for_display(display, wid)
        d.connect("realize", on_realize, parent)

        res = ""
        response = d.run()
        if response == Gtk.ResponseType.OK:
            res = d.get_filename()
            xpath = source2collection(res)
            if not os.path.exists(xpath):
                db = xapian.WritableDatabase(xpath, xapian.DB_CREATE)
                db.close()

            self.reload_all_collections()

        d.destroy()
        self.ChooserResponse(res)
        return

    #~ def display_image_chooser(self, parent_tag=""):
        #~ d = Gtk.FileChooserDialog(
            #~ title="Select an image",
            #~ parent=None,
            #~ action=gtk.FILE_CHOOSER_ACTION_OPEN,
            #~ buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                     #~ gtk.STOCK_OPEN,gtk.RESPONSE_OK))
#~ 
        #~ f = Gtk.FileFilter()
        #~ f.set_name("Supported Image Formats")
        #~ f.add_mime_type("image/png")
        #~ f.add_mime_type("image/jpeg")
        #~ f.add_pattern("*.png")
        #~ f.add_pattern("*.jpg")
        #~ d.add_filter(f)
#~ 
        #~ f = Gtk.FileFilter()
        #~ f.set_name("All files")
        #~ f.add_pattern("*")
        #~ d.add_filter(f)
#~ 
        #~ return self._display_dialog(d, parent_tag)

    @dbus.service.method('org.tw.XCollectionManager', in_signature='is', out_signature='')
    def display_gtk_chooser_dialog(self, wid, mode):
        GLib.idle_add(self._display_dialog, int(wid), str(mode))
        return 
