import os
import string
import threading

from gi.repository import GLib

from PyQt4.QtCore import (
    QAbstractListModel, QModelIndex, pyqtProperty, pyqtSignal)

from totallywired.utils.db import (
    get_album_qterm,
    get_doc_from_qterm,
    get_artist_qterm,
    qterm_in_db)

from totallywired.models.metadata import (
    ArtLoader,
    try_get_album_art)

from totallywired.db.collectors.enums import DbEnums
from totallywired.utils.lastfm import get_similar_artists, Similar


ALL_ARTIST_QUERY = "<b>All Artists</b>"


class SimilarArtistsFetcher(threading.Thread):

    def __init__(self, artist, callback):
        threading.Thread.__init__(self)
        self.daemon = True
        self.artist = artist
        self.callback = callback
        return

    def run(self):
        if callable(self.callback):
            self.callback(self.artist, get_similar_artists(artist=self.artist))
        return


class CurrentArtist(object):
    
    def __init__(self, artist):
        self.name = artist
        return


class Album(object):
    
    def __init__(self, facet):
        self._facet = facet
        self._title = ""
        return

    def _readdoc(self, db):
        doc, _ = get_doc_from_qterm(db, get_album_qterm(self._facet.term))
        if doc:
            self._title = doc.get_value(DbEnums.V_AARTIST)
            self._path = GLib.filename_from_uri(doc.get_data(), "")
        else:
            self._title = ""
            self._path = ""
        return

    def artist(self, db):
        if not self._title:
            self._readdoc(db)
        return self._title

    def path(self, db):
        if not self._path:
            self._readdoc(db)
        return self._path

    @property
    def name(self):
        return self._facet.term

    @property
    def termfreq(self):
        return self._facet.termfreq


class _AllFacet(object):

    def __init__(self, vals, artist):
        self.term = artist
        self.termfreq = 0
        self.similar = ""
        return


class AlbumsModel(QAbstractListModel):

    ROLE_ALBUM_NAME = "_album"
    ROLE_ALBUM_ART = "_art"
    ROLE_N_TRACKS = "_ntracks"
    ROLE_ARTIST = "_artist"

    COLUMNS = (
        ROLE_ALBUM_NAME,
        ROLE_ALBUM_ART,
        ROLE_N_TRACKS,
        ROLE_ARTIST,)

    ROLES = dict(enumerate(COLUMNS))

    _silentSignal = pyqtSignal()
    artReady = pyqtSignal(int)
    similarChanged = pyqtSignal()
    newArtist = pyqtSignal()

    _art_queue = []
    _art_loader_active_count = 0

    def __init__(self):
        QAbstractListModel.__init__(self)
        self.setRoleNames(AlbumsModel.ROLES)
        self._count = 0
        self._values = []
        self._job = ""
        self._title = ""
        self._similar = ""
        self._similar_list = []
        self._similar_available = False
        self._coll = None
        self.artReady.connect(self._update_art)
        return

    def _similar_ready(self, artist, similar):
        if similar and artist == self._title:
            self._similar = "<b>Similar to</b>: %s" % string.join((s.namelink for s in similar), ", ")
            similar.extend([CurrentArtist(artist),])
            self._similar_list = similar
            self._similar_available = True
        else:
            L = len(self._values)
            self._similar_available = False
            if L == 1:
                self._similar = "Displaying %s album" % L
            else:
                self._similar = "Displaying %s albums" % L
        self.similarChanged.emit()
        return

    def get_similar_list(self):
        return self._similar_list

    def clear(self):
        self.beginResetModel()
        self._count = 0
        self._values = []
        self._title = ""
        self._similar = ""
        self._similar_list = []
        self._similar_available = False
        self.endResetModel()
        return

    def set_spy(self, collection, spy, artist="", query=""):
        if (query or artist) == self._job:
            return

        self.clear()

        self._job = artist or query
        self._coll = collection
        vals = [Album(facet) for facet in list(spy.values())]
        self._count = len(vals)
        self.beginInsertRows(QModelIndex(), 0, self._count-1)
        self._values = vals
        self.endInsertRows()

        s_query = str(query)
        if s_query == "Xapian::Query(<alldocuments>)" or artist == "<b>All Artists</b>":
            self._title = "All Albums"
            self._similar = "Displaying %s albums" % len(self._values)
        elif query and not artist:
            self._title = "Matching albums..."
            self._similar = "Displaying %s albums" % len(self._values)
        else:
            self._title = artist
            job = SimilarArtistsFetcher(artist, self._similar_ready)
            job.start()

        self.newArtist.emit()
        self.similarChanged.emit()
        return

    def _update_art(self, i):
        self.dataChanged.emit(
            self.createIndex(i, 0),
            self.createIndex(i, 0))
        return

    def _art_loaded(self, path, index):
        if path:
            self.artReady.emit(index)

        self._art_loader_active_count -= 1
        print("ArtLoader job done... Remaining threads:%s, Queued:%s" % (self._art_loader_active_count, len(self._art_queue)))
        if self._art_queue:
            self._run_art_loader(*self._art_queue.pop(0))
            print("Queued job now running...")
        return

    def _run_art_loader(self, facet, artist, index):
        if self._art_loader_active_count > 8:
            self._art_queue.append((facet, artist, index))
            print("Max number of art loaders operating... adding job to queue.", facet.name)
            return os.path.join(os.getcwd(), "data/art/none.svg")

        loader = ArtLoader(
            facet.path(self._coll.db),
            facet.name,
            artist,
            "",                 # title, not reqd here
            self._art_loaded,   # callback
            index.row())        # callback args...

        self._art_loader_active_count += 1
        loader.start()
        return os.path.join(os.getcwd(), "data/art/none.svg")

    # QAbstractListModel compliance code
    def rowCount(self, parent=QModelIndex()):
        return self._count

    def data(self, index, role, roles=ROLES):
        if not index.isValid():
            return None

        rolev = roles[role]
        facet = self._values[index.row()]

        if rolev == self.ROLE_ALBUM_NAME:
            return facet.name.decode("utf-8", "ignore")
        elif rolev == self.ROLE_N_TRACKS:
            return int(facet.termfreq)
        elif rolev == self.ROLE_ARTIST:
            return facet.artist(self._coll.db).decode("utf-8", "ignore")
        elif rolev == self.ROLE_ALBUM_ART:
            artist = facet.artist(self._coll.db)
            art = try_get_album_art(facet.name, artist)
            if not art:
                art = self._run_art_loader(facet, artist, index)
            return art
        return None

    @pyqtProperty(str, notify=newArtist)
    def title(self):
        return self._title

    @pyqtProperty(str, notify=similarChanged)
    def similarArtistsLinks(self):
        return self._similar

    #~ @pyqtProperty(bool, notify=similarChanged)
    #~ def similarArtistsAvailable(self):
        #~ return self._similar_available

    @pyqtProperty(bool, notify=similarChanged)
    def hasSimilarArtistsInCollection(self):
        if not self._similar_available:
            return False
        db = self._coll.db
        for s in self._similar_list:
            if qterm_in_db(db, get_artist_qterm(s.name)):
                return True
        return False
