from PyQt4.QtCore import (
    QObject,
    QAbstractListModel,
    QModelIndex,
    pyqtSlot,
    pyqtSignal,
    pyqtProperty)

from .metadata import Metadata
from totallywired.player.common import Track
from totallywired.db.client import LikeSwitchJob


class _NoneTypeDocument(object):

    def get_data(*args):
        return ""

    def get_value(*args):
        return ""



class TrackListModel(QAbstractListModel):

    ROLE_TRACKNUM = "_track_n"
    ROLE_TITLE= "_title"
    ROLE_ALBUM = "_album"
    ROLE_ARTIST = "_artist"
    ROLE_DURATION = "_duration"
    ROLE_TAGS = "_tags"
    ROLE_LIKED = "_liked"
    ROLE_ENQUEUED = "_enqueued"
    ROLE_SELECTED = "_selected"

    QUEUE_EVENT_ADDED = 0
    QUEUE_EVENT_REMOVED = 1

    COLUMNS = (ROLE_TRACKNUM, ROLE_TITLE, ROLE_ALBUM, ROLE_ARTIST,
               ROLE_LIKED, ROLE_DURATION, ROLE_TAGS, ROLE_ENQUEUED,
               ROLE_SELECTED)
    ROLES = dict(enumerate(COLUMNS))

    selectionChanged = pyqtSignal(int)
    currentStreamIndexChanged = pyqtSignal()

    def __init__(self, app, collection):
        QAbstractListModel.__init__(self)
        self.setRoleNames(TrackListModel.ROLES)
        self.app = app
        self.player = None
        self.collection = collection
        # basic model state
        self._count = 0
        self._played = set()
        self._data = []
        # selection
        self.selected_indices = [-1,]
        # current stream if in list
        self._current_stream_index = -1
        return

    def __len__(self):
        return self._count

    def queue_changed(self, docid, event):
        if len(self._data) > 0:
            index = self.get_index_from_docid(docid)
            self.update_row(index)
        return

    # QAbstractListModel compliance code
    def rowCount(self, parent=QModelIndex()):
        return self._count

    def data(self, index, role):
        return self.do_data(index, role, self.ROLES)

    # public
    def clear(self):
        self.beginResetModel()
        self._data = []
        self._played.clear()
        self._count = 0
        self.endResetModel()
        return

    def get_document(self, index):
        return self.do_get_document(index)

    def get_document_from_docid(self, docid):
        if docid <= 0:
            return _NoneTypeDocument()
        return self.collection.db.get_document(docid)

    def get_docid(self, index):
        return self.do_get_docid(index)

    def get_index_from_docid(self, targetid):
        return self.do_get_index_from_docid(targetid)

    def update_row(self, i):
        if not (-1 < i < len(self._data)):
            return
        self.dataChanged.emit(
            self.createIndex(i, 0),
            self.createIndex(i, 0))
        return

    def update(self):
        if self.player.track is None:
            return

        player = self.player
        cur_index = self._current_stream_index
        index = self.get_index_from_docid(player.track.docid)

        self.update_row(cur_index)
        self.update_row(index)

        self._current_stream_index = index
        self.currentStreamIndexChanged.emit()
        return

    def reload_document(self, docid):
        i = self.get_index_from_docid(docid)
        self.update_row(i)
        return

    # must be implemented
    def do_data(self, index, role, roles):
        print("You must implement", self.do_data)
        return None

    def do_get_document(self, index):
        """
        Takes an index value and if valid returns
        (xapian.Document doc, int docid).
        If not valid, returns (None, -1).
        """
        print("You must implement", self.do_get_document)
        return None, -1

    def do_get_docid(self, index):
        """
        Takes an index value and if valid returns
        (int docid).
        If not valid, returns -1.
        """
        print("You must implement", self.do_get_docid)
        return

    def do_get_index_from_docid(self, targetid):
        print("You must implement", self.do_get_index_from_docid)
        return

    def do_get_current_stream_index(self):
        print("You must implement", self.do_get_current_stream_index)
        return -1

    @pyqtSlot(result=QObject)
    def get_current_metadata(self):
        print(self, "Fetching metadata ...")
        doc, docid = self.get_document(self.selected_indices[-1])
        # Metadata needs to be stored as an attr so it doesnt get gc'd
        self._metadata = Metadata(self, doc, docid)
        return self._metadata

    # qml exposed slots and properties
    @pyqtProperty(int, notify=selectionChanged)
    def selection(self):
        return self.selected_indices[-1]

    @pyqtSlot()
    def select_all(self):
        cur_index = self.selection
        self.selected_indices = [i for i in xrange(self._count) if i != cur_index]
        self.selected_indices.append(cur_index)

        for i in self.selected_indices:
            self.update_row(i)

        self.selectionChanged.emit(cur_index)
        return

    @pyqtSlot(int, int)
    def update_selection(self, index, modifier=0):
        selection = self.selected_indices
        modifier = int(modifier)

        if selection[0] == -1:
            del selection[0]

        if not modifier:
            # begin with now unselected rows needing update
            rows_to_update = selection
            # update selection
            selection = [index,]
            self.selected_indices = selection
            # extend with new rows needing update
            rows_to_update.extend(selection)
        elif modifier == 67108864:  # cntrl
            if index in selection:
                i = selection.index(index)
                del selection[i]
            else:
                selection.append(index)
            rows_to_update = [index,]
        elif modifier == 33554432:   # shift
            sel = selection[-1]

            if index > sel:
                index += 1
                r = [i for i in range(sel, index) if i not in selection]
            else:
                r = [i for i in range(index, sel) if i not in selection]
                r.reverse()

            selection.extend(r)
            rows_to_update = r

        if rows_to_update:
            for i in rows_to_update:
                self.update_row(i)

        self.selectionChanged.emit(index)
        return

    @pyqtSlot(int, result=bool)
    def is_selected(self, index):
        return index in self.selected_indices

    @pyqtProperty(int, notify=currentStreamIndexChanged)
    def current_stream_index(self):
        return self._current_stream_index

    @pyqtSlot()
    def enqueue_selection(self):
        queue = self.player.queue
        docids = [self.get_docid(i) for i in self.selected_indices if i != -1]
        for docid in docids:
            queue.enqueue_docid(docid)
        return

    @pyqtSlot()
    def unqueue_selection(self):
        self.do_unqueue_selection()
        return

    @pyqtSlot()
    def like_selection(self, dislike=False):
        c = self.collection
        docids = [self.get_docid(i) for i in self.selected_indices if i != -1]

        def finished():
            c.db.reopen()
            for docid in docids:
                c.documentChanged.emit(docid)
            print("LIKESWITCH", docid)
            return

        uris = [c.db.get_document(docid).get_data() for docid in docids]
        job = LikeSwitchJob(dislike, uris, finished)
        job.start()
        return

    @pyqtSlot()
    def dislike_selection(self):
        return self.like_selection(dislike=True)

    @pyqtSlot(result=bool)
    def start_current_item(self):
        docid = self.get_docid(self.selection)
        res = self.player._start_track(Track.from_docid(self, docid))
        if self.player.shuffle:
            self._played.add(docid)
        return res
