import os
import urllib2
import threading

from .base import (
    pyqtSlot,
    pyqtSignal,
    pyqtProperty,
    QModelIndex,
    TrackListModel)

from totallywired.db.collectors.enums import DbEnums


class CddaModel(TrackListModel):

    releaseUpdated = pyqtSignal()
    isWorking = pyqtSignal()

    def __init__(self, app, queue, manager):
        TrackListModel.__init__(self, app, manager)
        self.queue = queue

        self.ignored = app.ignored
        self._cdda = False
        self._firstrun = True
        self._working = False

        self.reload()
        return

    def reload(self):
        self.clear()

        c = self._cdda = self.collection.get_cdda_collection()
        print "CddaCollection:", c
        if self.collection.db and c and c.live:
            mset = self.collection.partial_query("__cdda")
            self._count = len(mset)
            self.beginInsertRows(QModelIndex(), 0, self._count-1)
            self._data = list([m.docid for m in mset])
            self.endInsertRows()

        self.releaseUpdated.emit()
        return

    # doers
    def do_get_document(self, index):
        docid = self._data[index]
        return self.collection.db.get_document(docid), docid

    def do_unqueue_selection(self):
        queue = self.queue
        docids = [self.get_docid(i) for i in self.selected_indices if i != -1]
        for docid in docids:
            queue.unqueue_docid(docid)
        return

    def do_get_docid(self, index):
        return self._data[index]

    def do_get_index_from_docid(self, targetid):
        if targetid in self._data:
            return self._data.index(targetid)
        return -1

    def do_data(self, index, role, roles):
        if not index.isValid() or self._data is None:
            return None

        doc, docid = self.get_document(index.row())
        rolev = self.ROLES[role]

        if rolev == self.ROLE_TRACKNUM:
            return doc.get_value(DbEnums.V_TRACKNUM)
        elif rolev == self.ROLE_TITLE:
            return str(doc.get_value(DbEnums.V_TITLE)).decode("utf-8", "ignore")
        elif rolev == self.ROLE_ALBUM:
            return str(doc.get_value(DbEnums.V_ALBUM)).decode("utf-8", "ignore")
        elif rolev == self.ROLE_ARTIST:
            return str(doc.get_value(DbEnums.V_ARTIST)).decode("utf-8", "ignore")
        elif rolev == self.ROLE_DURATION:
            return doc.get_value(DbEnums.V_DURATION_FSTR)
        elif rolev == self.ROLE_TAGS:
            return doc.get_value(DbEnums.V_TAGS).decode("utf-8", "ignore")
        elif rolev == self.ROLE_ENQUEUED:
            return self.queue.has_docid(docid)
        elif rolev == self.ROLE_LIKED:
            return doc.get_value(DbEnums.V_LIKE) or '0'
        elif rolev == self.ROLE_SELECTED:
            return index.row() in self.selected_indices
        return None

    def do_get_current_stream_index(self):
        if self.player.track is None:
            return -1
        docid = self.player.track.docid
        if docid in self._data:
            return self._data.index(docid)
        return -1

    def collection_changed(self, i, collection):
        if collection.is_cdda:
            self._working = collection.working
            self.isWorking.emit()
        return

    @pyqtSlot(str, str)
    def release_lookup(self, album_q, artist_q):
        if self.ignored:
            return
        elif self._working:
            print("working...")
            return

        self._working = True
        self.isWorking.emit()

        q = '"%s" AND artist:"%s"' % (album_q, artist_q)
        self.collection.service.rerun_cdda_collector_with_mbquery(
            q, self._cdda.disc_num)
        return

    @pyqtSlot(int)
    def set_disc_number(self, n):
        if self._working:
            print("working...")
            return

        self._working = True
        self.isWorking.emit()

        print("Rerun cdda collector.  DIsc number:", n)
        self.collection.service.rerun_cdda_collector(int(n))
        return

    @pyqtSlot(result=int)
    def get_disc_number(self):
        if self._cdda:
            return self._cdda.disc_num
        return 0

    @pyqtProperty(int, notify=releaseUpdated)
    def disc_count(self):
        if self._cdda:
            return self._cdda.disc_count
        return 0

    @pyqtProperty(int, notify=releaseUpdated)
    def disc_number(self):
        if self._cdda:
            return self._cdda.disc_num
        return 0

    @pyqtProperty(str, notify=releaseUpdated)
    def album_art_url(self):
        def finished(*args):
            print("Caching finished!")

        if self._cdda:
            url = self._cdda.album_art
            if url:
                ext = os.path.splitext(url)[-1]
                localpath = os.path.join(self._cdda.xpath, "AlbumArt%s" % ext)
                if not os.path.exists(localpath):
                    # download and cache artwork
                    print("Caching url to", localpath)
                    cacher = ImageCacher(url, localpath, finished)
                    cacher.start()
                    return url
                else:
                    return localpath
        return ""

    @pyqtProperty(str, notify=releaseUpdated)
    def tags(self):
        if self._cdda:
            return self._cdda.tags
        return ""

    @pyqtProperty(str, notify=releaseUpdated)
    def date(self):
        if self._cdda:
            return self._cdda.date
        return ""

    @pyqtProperty(str, notify=releaseUpdated)
    def artist(self):
        if self._cdda:
            return self._cdda.artist
        return ""

    @pyqtProperty(str, notify=releaseUpdated)
    def album(self):
        if self._cdda:
            return self._cdda.album
        return ""

    @pyqtProperty(bool, notify=isWorking)
    def working(self):
        return self._working or self._cdda and self._cdda.working


class ImageCacher(threading.Thread):

    def __init__(self, url, filename, callback=False):
        threading.Thread.__init__(self)
        self.daemon = True
        self.url = url
        self.filename = filename
        self.callback = callback
        return

    def run(self):
        try:
            opener = urllib2.build_opener()
        except urllib2.URLError:
            return
        except Exception as e:
            print("Somthing bad happend: %s" % e)
            return

        art = opener.open(self.url)
        data = art.read()

        fout = open(self.filename, "wb")
        fout.write(data)
        fout.close()


        if self.callback:
            self.callback()
        return
