#
# Authors:
#  Michael Vogt
#  Ollivier Tilloy
#  Matthew McGowan
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from PyQt4.QtCore import QAbstractListModel, QModelIndex, pyqtSlot


class IndexModel(QAbstractListModel):

    ROLE_SOURCE = "_source"
    ROLE_INDEX = "_index"
    ROLE_STATE = "_state"
    ROLE_PROGRESS = "_progress"
    ROLE_SRC_TYPE = "_srctype"
    ROLE_SRC_NAME = "_srcname"
    ROLE_SRC_ICON = "_srcicon"
    ROLE_DOCCOUNT = "_doccount"
    ROLE_DISPL_TEXT = "_displ_text"

    COLUMNS = (ROLE_SOURCE, ROLE_INDEX, ROLE_STATE, ROLE_PROGRESS,
               ROLE_DOCCOUNT, ROLE_SRC_TYPE, ROLE_SRC_NAME,
               ROLE_SRC_ICON, ROLE_DISPL_TEXT)

    ROLES = dict(enumerate(COLUMNS))

    def __init__(self, collections):
        QAbstractListModel.__init__(self)
        self.setRoleNames(IndexModel.ROLES)
        self.collections = collections
        self._count = 0
        self.update_collections()
        collections.collectionAdded.connect(self.collection_added)
        return

    def _update_row(self, i):
        self.dataChanged.emit(
            self.createIndex(i, 0),
            self.createIndex(i, 0))
        return

    def collection_added(self, i, source):
        self._count += 1
        self.beginInsertRows(QModelIndex(), i, i)
        self.endInsertRows()
        return

    def collection_changed(self, i, collection):
        self._update_row(i)
        return

    def clear(self):
        self.beginResetModel()
        self._count = 0
        self.endResetModel()
        return

    def update_collections(self):
        self.clear()
        indexes = self.collections.list()
        self._count = len(indexes)
        self.beginInsertRows(QModelIndex(), 0, self.rowCount()-1)
        self.endInsertRows()
        return

    @pyqtSlot(int)
    def remove_index(self, i):
        self._count -= 1
        self.beginRemoveRows(QModelIndex(), i, i)
        self.endRemoveRows()
        return

    @pyqtSlot(str, str, result=bool)
    def state_is(self, state, substate):
        state = str(state)
        substates = str(substate).split("|")
        for s in substates:
            if s in state:
                return True
        return False

    # QAbstractListModel compliance code
    def rowCount(self, parent=QModelIndex()):
        return self._count

    def data(self, index, role, roles=ROLES):
        if not index.isValid():
            return None

        rolev = self.ROLES[role]
        collection = self.collections.list()[index.row()]

        if rolev == self.ROLE_SOURCE:
            if not collection.source:
                return "No inserted"
            return collection.source
        elif rolev == self.ROLE_INDEX:
            return collection.xpath
        elif rolev == self.ROLE_STATE:
            return collection.state
        elif rolev == self.ROLE_PROGRESS:
            return collection.progress
        elif rolev == self.ROLE_SRC_TYPE:
            if collection.removable:
                return "Removable Sources"
            return "Home Sources"
        elif rolev == self.ROLE_SRC_NAME:
            return collection.title
        elif rolev == self.ROLE_SRC_ICON:
            return collection.icon_source
        elif rolev == self.ROLE_DOCCOUNT:
            return collection.doccount
        elif rolev == self.ROLE_DISPL_TEXT:
            if collection.progress != -1:
                return "Working... %.2f%%" % round(collection.progress * 100, 2)
            elif collection.live:
                return "Active"
            elif collection.hidden:
                return "Hidden"
            return "Not connected"
        return None
