import os
import time
import fnmatch
import shutil
import hashlib
import threading

import Image
from StringIO import StringIO

from gi.repository import GLib

from PyQt4.QtCore import (
    QObject,
    QAbstractListModel,
    QModelIndex,
    pyqtSlot,
    pyqtSignal,
    pyqtProperty)

from totallywired.utils.caching import Caches
from totallywired.db.client import MetadataUpdateJob
from totallywired.db.collectors import get_metadata
from totallywired.db.collectors.enums import DbEnums
from totallywired.utils.misc import remove_diacritical_marks
from totallywired.utils.lastfm import (
    query_artwork, 
    get_artist_biography,
    get_similar_tracks)


def get_potential_art_paths(album, artist, title=""):
    if not isinstance(artist, str):
        artist = artist.encode('unicode_escape')
    if not isinstance(album, str):
        album = album.encode('unicode_escape')
    h = hashlib.md5()
    h.update(artist.lower() + album.lower())
    s = h.hexdigest()
    aart = os.path.join(Caches.ALBUMART, "%s.png" % s)
    allart = os.path.join(Caches.ALBUMART, "%s.all.png" % s)
    if title:
        h.update(title.lower().encode('unicode_escape'))
        title = h.hexdigest()
        title = os.path.join(Caches.ALBUMART, "%s.%s" % (s, title))
    return title, aart, allart

def try_get_album_art(album, artist):
    _, aart, allart = get_potential_art_paths(album, artist)
    if os.path.exists(allart):
        return allart
    elif os.path.exists(aart):
        return aart
    return ""


_htmlstyle = "<html><style type='text/css'>a:link {text-decoration:none;color:#252525;} * {line-height:140%%;}</style><body>%s</body></html>"
class BioFetcher(threading.Thread):

    def __init__(self, mbid="", artist="", lang="en", callback=None):
        threading.Thread.__init__(self)
        self.daemon = True
        self.mbid = mbid
        self.artist = artist
        self.lang = lang
        self.callback = callback
        return

    def run(self):
        if callable(self.callback):
            bio = get_artist_biography(artist=self.artist)
            if bio:
                self.callback(_htmlstyle % bio)
                return
            self.callback("")
        return


_web_attempted = set()
class ArtLoader(threading.Thread):

    SIZE = 300

    KNOWN_ALBUM_ART_KEYS = (
        'APIC:cover',
        'APIC:',
        'APIC:Folder.jpg',
        'covr')

    def __init__(self, path, album, artist, title, finished_cb, *args):
        threading.Thread.__init__(self)
        self.daemon = True
        self.path = path
        self.album = album
        self.artist = artist
        self.title = title
        self.data = None
        self.finished_cb = finished_cb
        self.args = args
        return

    def __str__(self):
        return "ArtLoader(%s, %s, %s)" % (self.album, self.artist, self.title)

    def cleanup_track_art(self, art, aart, allart):
        # create special file
        shutil.copyfile(aart, allart)
        # clean up old files
        t = os.path.splitext(os.path.split(aart)[1])[0]
        mset = fnmatch.filter(os.listdir(Caches.ALBUMART), t+"*")
        if len(mset) >= 2:
            for m in mset:
                path = os.path.join(Caches.ALBUMART, m)
                if path == allart:
                    continue
                os.remove(path)
        return

    def is_similar_image(self, im1, im2):
        im1.thumbnail((1, 1), Image.ANTIALIAS)
        im2.thumbnail((1, 1), Image.ANTIALIAS)
        p1 = im1.getpixel((0,0))
        p2 = im2.getpixel((0,0))
        # tolerate a small level of difference
        if not isinstance(p1, tuple) or not isinstance(p2, tuple):
            return False
        similar = (
            p2[0]-2 <= p1[0] <= p2[0]+2 and
            p2[1]-2 <= p1[1] <= p2[1]+2 and
            p2[2]-2 <= p1[2] <= p2[2]+2)
        return similar

    def cache_image(self, cache_path, image=None, image_path=""):
        if image_path:
            try:
                im = Image.open(image_path)
            except IOError as e:
                print("Could not open image. %s" % e)
                from totallywired.utils.icons import IconFetcher
                fchr = IconFetcher("mimetypes")
                err_im = fchr.get_iconpath64("unknown")
                im = Image.open(err_im)
                im.save(cache_path, "PNG")
                return
        else:
            im = image
        w, h = im.size
        # downscale image if it is huuuuuuge
        if max(w, h) > self.SIZE:
            im.thumbnail((self.SIZE, self.SIZE), Image.ANTIALIAS)
        elif max(w, h) < (self.SIZE - 100):
            # try for better quality online
            success = self.get_online_art(cache_path)
            if success:
                del im
                return
        print("Saving art to: %s" % cache_path)
        im.save(cache_path, "PNG")
        return

    def get_online_art(self, cache_path):
        global _web_attempted
        if cache_path in _web_attempted:
            print("A web attempt as already been made for %s-%s.  Skipping..." % (self.album, self.artist))
            return False

        _web_attempted.add(os.path.split(cache_path)[-1])
        artwork_url = query_artwork(self.album, self.artist)
        if not artwork_url:
            print("No art could be found!")
            return False
        # retrieve and cache file
        import urllib, time
        tmp = os.path.join(Caches.TEMP, str(time.time()))
        urllib.urlretrieve(artwork_url, tmp)
        self.cache_image(cache_path, image_path=tmp)
        return cache_path

    def get_embedded_art(self, art, aart, allart):
        data, _ = get_metadata(self.path)
        if data is not None:
            for k in self.KNOWN_ALBUM_ART_KEYS:
                if data.has_key(k):
                    try:
                        im = Image.open(StringIO(data[k].data))
                    except Exception, e:
                        print("An exception occurred: %s" % e)
                        continue

                    if not os.path.exists(aart):
                        self.cache_image(aart, image=im)
                        similar = False
                    else:
                        # compare images and see if they are similar
                        similar = self.is_similar_image(im.copy(), Image.open(aart))
                    if not similar and art:
                        self.cache_image(art, image=im)
                        return art
                    elif art:
                        self.cleanup_track_art(art, aart, allart)
                    else:
                        return aart
                    return allart
        return False

    def get_folder_art(self, aart):
        d = os.path.split(self.path)[0]
        if not os.path.exists(d):
            return False
        imgs = fnmatch.filter(os.listdir(d), "*.jpg") + fnmatch.filter(os.listdir(d), "*.png")
        if not imgs:
            return False
        elif "Folder.jpg" in imgs:
            f = "Folder.jpg"
        elif "Cover.jpg" in imgs:
            f = "Cover.jpg"
        else:
            f = imgs[0]
        self.cache_image(aart, image_path=os.path.join(d, f))
        return aart

    def run(self):
        art, aart, allart = get_potential_art_paths(self.album, self.artist, self.title)
        res = False
        if not res:
            if os.path.exists(art):
                res = art
        if not res:
            if os.path.exists(allart):
                res = allart
        if not res:
            res = self.get_embedded_art(art, aart, allart)
        if not res:
            if os.path.exists(aart):
                res = aart
        if not res:
            res = self.get_folder_art(aart)
        if not res:
            res = self.get_online_art(aart)

        self.finished_cb(res, *self.args)
        # i do this to ensure we throttle our usage of lastfm to a very 
        # safe level...
        time.sleep(0.2)
        return


class Metadata(QObject):

    QML_FIELD_DATA_OBJ_NAMES = (
        "title",
        "album",
        "ensemble",
        "artist",
        "date",
        "tracknumber",
        "tags")

    metadataChanged = pyqtSignal()
    albumArtChanged = pyqtSignal()
    selectionChanged = pyqtSignal()
    bioChanged = pyqtSignal()

    def __init__(self, model, doc, docid):
        QObject.__init__(self)
        self.model = model
        self._index = model.selection
        self.__index = self._index
        self._new(doc, docid)
        return

    def _new(self, doc, docid):
        self.doc = doc
        self.docid = docid
        self._art = ""
        self._bio = ""
        # start async jobs
        loader = ArtLoader(
            GLib.filename_from_uri(doc.get_data(), ""),
            doc.get_value(DbEnums.V_ALBUM),
            doc.get_value(DbEnums.V_ARTIST),
            doc.get_value(DbEnums.V_TITLE),
            self._art_loaded)
        loader.start()
        loader = BioFetcher(
            artist=doc.get_value(DbEnums.V_ARTIST),
            callback=self._bio_loaded)
        loader.start()
        return

    def _art_loaded(self, path):
        if path:
            self._art = path
            self.albumArtChanged.emit()
        return

    def _bio_loaded(self, bio):
        if bio:
            self._bio = bio
            self.bioChanged.emit()
        return

    @pyqtProperty(str, notify=bioChanged)
    def bio(self):
        return self._bio

    @pyqtProperty(str, notify=metadataChanged)
    def trackn(self):
        return self.doc.get_value(DbEnums.V_TRACKNUM)

    @pyqtProperty(str, notify=metadataChanged)
    def title(self):
        return self.doc.get_value(DbEnums.V_TITLE).decode("utf-8", "ignore")

    @pyqtProperty(str, notify=metadataChanged)
    def album(self):
        return self.doc.get_value(DbEnums.V_ALBUM).decode("utf-8", "ignore")

    @pyqtProperty(str, notify=metadataChanged)
    def artist(self):
        return self.doc.get_value(DbEnums.V_ARTIST).decode("utf-8", "ignore")

    @pyqtProperty(str, notify=metadataChanged)
    def aartist(self):
        return self.doc.get_value(DbEnums.V_AARTIST).decode("utf-8", "ignore")

    @pyqtProperty(str, notify=metadataChanged)
    def duration_fstr(self):
        return self.doc.get_value(DbEnums.V_DURATION_FSTR)

    @pyqtProperty(str, notify=metadataChanged)
    def tags(self):
        return self.doc.get_value(DbEnums.V_TAGS).decode("utf-8", "ignore")

    @pyqtProperty(str, notify=metadataChanged)
    def date(self):
        return self.doc.get_value(DbEnums.V_DATE)

    @pyqtProperty(str, notify=metadataChanged)
    def path(self):
        return GLib.filename_from_uri(self.doc.get_data(), "").decode("utf-8", "ignore")

    @pyqtProperty(str, notify=metadataChanged)
    def document_id(self):
        return str(self.docid)

    @pyqtProperty(str, notify=albumArtChanged)
    def album_art(self):
        return self._art

    @pyqtProperty(bool, notify=selectionChanged)
    def multi_select(self):
        return len(self.model.selected_indices) > 1

    def _shift(self, step):
        model = self.model
        if model._count <= 1:
            return

        if len(model.selected_indices) > 1:
            # a range selected
            sel = model.selected_indices
            i = sel.index(self._index)

            if sel[0] > sel[-1]:   # selection direction ascending
                step *= -1

            i += step
            if i < 0:
                i = len(model.selected_indices) - 1
            elif i >= len(model.selected_indices):
                i = 0

            self.__index = i
            self._index = model.selected_indices[i]
        else:
            self._index += step

            if self._index >= model._count:
                self._index = 0
            elif self._index < 0:
                self._index = model._count - 1

            self.__index = self._index
            model.update_selection(self._index)

        self._new(*self.model.get_document(self._index))
        self.metadataChanged.emit()
        self.albumArtChanged.emit()
        return

    @pyqtSlot()
    def request_next(self):
        return self._shift(step=1)

    @pyqtSlot()
    def request_previous(self):
        return self._shift(step=-1)


    @pyqtSlot(str, str)
    def field_save_changes_to_selection(self, field, change):

        def finished(docids):
            c = self.model.collection
            c.db.reopen()
            for docid in docids:
                c.documentChanged.emit(docid)
            print("Changes saved")
            self.metadataChanged.emit()
            return

        model = self.model
        change_map = {}
        docids = []
        for doc, docid in (model.get_document(i) for i in model.selected_indices):
            changes = {
                DbEnums.T_TITLE: doc.get_value(DbEnums.V_TITLE),
                DbEnums.T_ALBUM: doc.get_value(DbEnums.V_ALBUM),
                DbEnums.T_ARTIST: doc.get_value(DbEnums.V_ARTIST),
                DbEnums.T_AARTIST: doc.get_value(DbEnums.V_AARTIST),
                DbEnums.T_TRACKNUM: doc.get_value(DbEnums.V_TRACKNUM),
                DbEnums.T_DATE: doc.get_value(DbEnums.V_DATE),
                DbEnums.T_TAGS: doc.get_value(DbEnums.V_TAGS),}

            changes[str(field)] = str(change)
            change_map[doc.get_data()] = changes
            docids.append(docid)

        job = MetadataUpdateJob(
            change_map,
            finished,       # callback
            docids)   # args
        job.start()        
        return

    @pyqtSlot()
    def save_changes(self):

        def finished():
            c = self.model.collection
            c.db.reopen()
            c.documentChanged.emit(self.docid)
            print("Changes saved")
            self.metadataChanged.emit()
            return

        changes = {}

        root = self.model.app.view.rootObject()
        editor = root.findChild(QObject, "editor")

        for oname in self.QML_FIELD_DATA_OBJ_NAMES:
            obj = editor.findChild(QObject, oname)
            changes[oname] = unicode(obj.get_changes().toString())

        print("Metadata changes to be saved:")
        print(changes)

        if changes:
            change_map = {self.doc.get_data(): changes,}
            job = MetadataUpdateJob(
                change_map,
                finished)   # finished cb
            job.start()
        return
