#
# Authors:
#  Michael Vogt
#  Ollivier Tilloy
#  Matthew McGowan
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import re
import xapian
import random

from .base import (
    pyqtSlot,
    pyqtSignal,
    pyqtProperty,
    QModelIndex,
    TrackListModel)

from totallywired.db.collectors.enums import DbEnums
from totallywired.models.artists import ArtistsModel
from totallywired.models.cdda import CddaModel
from totallywired.models.albums import AlbumsModel, ALL_ARTIST_QUERY

from totallywired.utils.misc import (
    SubMany, 
    remove_diacritical_marks)

from totallywired.utils.db import (
    get_qterm,
    get_artist_qterm,
    get_album_qterm,
    qterm_in_db,
    get_docids_from_group_qterm,
    is_alldocuments_query)


XAPIAN_ALL_QUERY = "Xapian::Query(<alldocuments>)"


class MusicCollection(TrackListModel):

    _silentSignal = pyqtSignal()
    searchChanged = pyqtSignal()
    currentArtistChanged = pyqtSignal()

    def __init__(self, app, queue, collection):
        TrackListModel.__init__(self, app, collection)

        self._subber = SubMany({
            'DISLIKED': 'lkd:hate',
            'LIKED': 'lkd:like',
            '><': 'lkd:hate',
            '^^': 'lkd:like',
            '&': ' AND ',
            '|': ' OR ',
            '!': ' NOT '})

        self._cur_str_q = xapian.Query("")
        self._cur_art_q = ""
        self._cur_alb_q = ""
        # keep a set of the data so we needn't recreate a set on each
        # shuffle request
        self._dataset = set()
        self._artists = ArtistsModel()
        self._albums = AlbumsModel()
        self._cdda = CddaModel(app, queue, collection)
        self._currentArtist = 0
        self.queue = queue
        return

    # internal stuff
    def __getitem__(self, i):
        return self._data[i]

    def _update_for_new_mset(self, mset):
        if len(mset):
            self._count = len(mset)
            self.beginInsertRows(QModelIndex(), 0, self._count-1)
            self._data = list([m.docid for m in mset])
            self.endInsertRows()

        self.update()
        self.update_selection(0)
        self.currentStreamIndexChanged.emit()

        # store a data set
        self._dataset = set(self._data)
        self.searchChanged.emit()
        return

    def _preparse_qstr(self, q):
        q = self._subber.sub(q)

        if "artist(" in q:
            res = re.findall(r"artist\((.+)\)", q)
            if res:
                artist_ndmt = remove_diacritical_marks(res[0])
                artist_qterm = get_artist_qterm(artist_ndmt)
                q = q.replace("artist(%s)" % res[0], artist_qterm)
            else:
                return q

        elif "album(" in q:
            res = re.findall(r"album\((.+)\)", q)
            if res:
                album_ndmt = remove_diacritical_marks(res[0])
                album_qterm = get_album_qterm(album_ndmt)
                q = q.replace("album(%s)" % res[0], album_qterm)
            else:
                return q
        return q

    # doers
    def do_get_document(self, index):
        docid = self._data[index]
        return self._db.get_document(docid), docid

    def do_unqueue_selection(self):
        queue = self.player.queue
        docids = [self.get_docid(i) for i in self.selected_indices if i != -1]
        for docid in docids:
            queue.unqueue_docid(docid)
        return

    def do_get_docid(self, index):
        return self._data[index]

    def do_get_index_from_docid(self, targetid):
        if targetid in self._data:
            return self._data.index(targetid)
        return -1

    def do_data(self, index, role, roles):
        if not index.isValid() or self._data is None:
            return None

        doc, docid = self.get_document(index.row())
        rolev = self.ROLES[role]

        if rolev == self.ROLE_TRACKNUM:
            return doc.get_value(DbEnums.V_TRACKNUM)
        elif rolev == self.ROLE_TITLE:
            return str(doc.get_value(DbEnums.V_TITLE)).decode("utf-8", "ignore")
        elif rolev == self.ROLE_ALBUM:
            return str(doc.get_value(DbEnums.V_ALBUM)).decode("utf-8", "ignore")
        elif rolev == self.ROLE_ARTIST:
            return str(doc.get_value(DbEnums.V_ARTIST)).decode("utf-8", "ignore")
        elif rolev == self.ROLE_DURATION:
            return doc.get_value(DbEnums.V_DURATION_FSTR)
        elif rolev == self.ROLE_TAGS:
            return doc.get_value(DbEnums.V_TAGS).decode("utf-8", "ignore")
        elif rolev == self.ROLE_ENQUEUED:
            return self.player.queue.has_docid(docid)
        elif rolev == self.ROLE_LIKED:
            return doc.get_value(DbEnums.V_LIKE) or '0'
        elif rolev == self.ROLE_SELECTED:
            return index.row() in self.selected_indices
        return None

    def do_get_current_stream_index(self):
        if self.player.track is None:
            return -1
        docid = self.player.track.docid
        if docid in self._data:
            return self._data.index(docid)
        return -1

    def clear(self):
        self._dataset.clear()
        return TrackListModel.clear(self)

    # slots
    @pyqtProperty(ArtistsModel, notify=_silentSignal)
    def artists(self):
        return self._artists

    @pyqtProperty(AlbumsModel, notify=_silentSignal)
    def albums(self):
        return self._albums

    @pyqtProperty(CddaModel, notify=_silentSignal)
    def cdda(self):
        return self._cdda

    @pyqtSlot(str)
    def new_query_from_string(self, qstr):
        qstr = self._preparse_qstr(str(qstr))
        if qstr:
            try:
                q = self._qparser.parse_query(
                    qstr,
                    self.collection.DEFAULT_QUERY_FLAGS)
            except Exception as e:
                print('Ignoring malformed query!\n  ', e)
                return
        else:
            q = xapian.Query("")

        self._cur_str_q = q
        self._cur_art_q = ""
        return self.new_query(self._cur_str_q)

    @pyqtSlot()
    def reset_current_artist_query(self):
        if not self._cur_alb_q:
            print("no album query...")
            return
        self._cur_alb_q = ""
        if not is_alldocuments_query(self._cur_str_q) and self._cur_art_q:
            q = xapian.Query(
                xapian.Query.OP_AND,
                self._cur_art_q,
                self._cur_str_q)
        else:
            q = self._cur_art_q

        self.clear()
        self._enq.set_query(q)
        self._enq.set_sort_by_value(DbEnums.V_SORTING, False)
        mset = self._enq.get_mset(0, self._db.get_doccount())
        self._update_for_new_mset(mset)
        return

    @pyqtSlot(str)
    def filter_current_query_by_artist(self, artist_name, artist_spy=False, highlight_artist=False):
        self.clear()
        artist_name = unicode(artist_name)
        allartist = artist_name == ALL_ARTIST_QUERY

        if allartist:
            q = self._cur_str_q
            self._cur_art_q = ""
            spy = xapian.ValueCountMatchSpy(DbEnums.V_ARTIST)
            self._enq.add_matchspy(spy)
        else:
            artist_name_ndmt = remove_diacritical_marks(artist_name)
            self._cur_art_q = xapian.Query(get_artist_qterm(artist_name_ndmt))
            self._cur_alb_q = ""
            q = xapian.Query(
                xapian.Query.OP_AND,
                self._cur_art_q,
                self._cur_str_q,)

        spy = xapian.ValueCountMatchSpy(DbEnums.V_ALBUM)
        self._enq.clear_matchspies()
        self._enq.add_matchspy(spy)
        if artist_spy:
            ar_spy = xapian.ValueCountMatchSpy(DbEnums.V_AARTIST)
            self._enq.add_matchspy(ar_spy)

        self._enq.set_query(q)
        self._enq.set_sort_by_value(DbEnums.V_SORTING, False)

        mset = self._enq.get_mset(0, self._db.get_doccount())

        self._albums.set_spy(self.collection, spy, artist=artist_name)
        if artist_spy:
            self._artists.set_spy(ar_spy)
        if highlight_artist:
            if allartist:
                self._currentArtist = 0
            else:
                self._currentArtist = self._artists.get_artist_index(artist_name)
            self.currentArtistChanged.emit()

        self._update_for_new_mset(mset)
        return

    #~ @pyqtSlot()
    #~ def reset_artist_filter(self):
        #~ self.clear()
        #~ q = self._cur_str_q
        #~ self._cur_art_q = ""
        #~ spy = xapian.ValueCountMatchSpy(DbEnums.V_ARTIST)
        #~ self._enq.clear_matchspies()
        #~ self._enq.add_matchspy(spy)
        #~ self._enq.set_query(q)
        #~ self._enq.set_sort_by_value(DbEnums.V_SORTING, False)
        #~ mset = self._enq.get_mset(0, self._db.get_doccount())
        #~ self._artists.set_spy(spy)
        #~ self._update_for_new_mset(mset)
        #~ return

    @pyqtProperty(int, notify=currentArtistChanged)
    def current_artist_index(self):
        return self._currentArtist

    @pyqtProperty(bool, notify=searchChanged)
    def artist_filter_active(self):
        return self._cur_art_q != ""

    @pyqtSlot(str)
    def filter_current_query_by_album(self, album_name):
        self.clear()
        album_name = unicode(album_name)
        allalbum = album_name.startswith("<b>")

        if allalbum:
            self._cur_alb_q = xapian.Query("")
            q = xapian.Query(
                xapian.Query.OP_AND,
                self._cur_str_q,
                self._cur_art_q)
        else:
            album_name_ndmt = remove_diacritical_marks(album_name)
            self._cur_alb_q = xapian.Query(get_album_qterm(album_name_ndmt))
            q = xapian.Query(
                xapian.Query.OP_AND,
                self._cur_alb_q,
                self._cur_str_q)

        self._enq.set_query(q)
        self._enq.set_sort_by_value(DbEnums.V_SORTING, False)
        mset = self._enq.get_mset(0, self._db.get_doccount())
        self._update_for_new_mset(mset)
        return

    # other public methods but so far not exposed to the qml
    def reload(self):
        self._data = []
        self._count = 0

        self._db = self.collection.db
        self._enq = self.collection._enq
        self._qparser = self.collection._parser
        doccount = 0
        # unset the cached artist param in the album model so that it 
        # reloads
        self._albums._artist = ""
        if self._db:
            self.new_query(self._cur_str_q)
            doccount = self._db.get_doccount()
            self.player.sync_with_remote_player()
        else:
            self.clear()
        print("Reloaded collection.  Searching %s docs." % doccount)

        self._cdda.reload()
        return

    def reload_model(self):
        # update db
        self.reload()
        self.player.update_docids()
        self.currentStreamIndexChanged.emit()
        return

    def reload_document(self, docid):
        TrackListModel.reload_document(self, docid)
        if self.collection.cdda_present:
            self._cdda.reload_document(docid)
        return

    def new_query(self, q):
        if not self._db:
            return

        self.clear()
        self._enq.set_query(q)
        self._enq.set_sort_by_value(DbEnums.V_SORTING, False)

        self._enq.clear_matchspies()

        ar_spy = xapian.ValueCountMatchSpy(DbEnums.V_AARTIST)
        self._enq.add_matchspy(ar_spy)

        al_spy = xapian.ValueCountMatchSpy(DbEnums.V_ALBUM)
        self._enq.add_matchspy(al_spy)

        mset = self._enq.get_mset(0, self._db.get_doccount())

        self._artists.set_spy(ar_spy)
        self._albums.set_spy(self.collection, al_spy, query=q)
        self._update_for_new_mset(mset)
        return

    def next_shuffled(self, excluded_docids):
        print("ExcludedDocids: %s" % len(self._played))
        if len(excluded_docids) == self._count:
            if self.player.repeat:
                print("All tracks have been played once. Repeat is ON; exluded_docids will now be reset and playback will continue...")
                excluded_docids.clear()
            else:
                print("All tracks have been played once. Repeat is OFF; playback will now stop.")
                return (False, -1)

        candidates = self._dataset - excluded_docids

        if len(candidates):
            picked = random.sample(candidates, 1)[0]
        else:
            picked = False

        if picked and self._docid_is_hated(picked):
            excluded_docids.add(picked)
            return self.next_shuffled(excluded_docids)
        elif picked:
            index = self._data.index(picked)
        else:
            index = -1
        return (picked, index)

    def next_one(self):
        index = self._current_stream_index
        # check and bump index value
        if index == -1:
            if not self._data:
                return (False, -1)
            index = 0
        else:
            index += 1

        # check new index is within range
        if index >= self._count:
            # if new index exceeds range and repeat is ON, return to 0
            if self.player.repeat:
                index = 0
                picked = self._data[0]
                print("The end of the tracklist has been reached. Repeat is ON; playback will resume from the top...")
            else:
                print("The end of the tracklist has been reached. Repeat is OFF; playback will now stop.")
                return (False, -1)
        else:
            picked = self._data[index]
            if self._docid_is_hated(picked):
                self._current_stream_index += 1
                return self.next_one()

        return (picked, index)

    def get_next(self):
        player = self.player

        if self.queue.has_tracks():
            docid = self.queue.next_track().docid
            index = None
        elif not player.shuffle:
            docid, index = self.next_one()
        elif player.shuffle:
            docid, index = self.next_shuffled(self._played)
        if docid:
            self._played.add(docid)
        return docid, index

    def _docid_is_hated(self, docid):
        v = self._db.get_document(docid).get_value(DbEnums.V_LIKE)
        return v == "2"

    def get_docid_from_uri(self, uri):
        if not self._db:
            return -1
        pl = self._db.postlist(get_qterm(uri))
        try:
            plitem = next(pl)
        except StopIteration:
            raise KeyError("No document with QTerm: %s" % get_qterm(uri))
            return -1
        return plitem.docid

    @pyqtSlot(str, result=bool)
    def handle_link(self, link):
        artist = str(link).split("/")[-1].replace("+", " ")
        qt = get_artist_qterm(artist)
        if qterm_in_db(self._db, qt):
            self._cur_str_q = xapian.Query("")
            self.filter_current_query_by_artist(artist, highlight_artist=True)
            return True
        return False

    @pyqtSlot()
    def generate_similar_playlist(self):
        playlist = []
        for s in self._albums.get_similar_list():
            qt = get_artist_qterm(s.name)
            playlist.extend(get_docids_from_group_qterm(self._db, qt))

        # always random...?
        random.shuffle(playlist)

        print("Random playlist of similar artists generated! Length:%s" % len(playlist))
        queue = self.queue
        for docid in playlist:
            queue.enqueue_docid(docid)
        return
