#
# Authors:
#  Michael Vogt
#  Ollivier Tilloy
#  Matthew McGowan
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .base import (
    pyqtSlot,
    pyqtSignal,
    pyqtProperty,
    QModelIndex,
    Track,
    TrackListModel)

from totallywired.player.common import Queue
from totallywired.db.collectors.enums import DbEnums


class QueueModel(Queue, TrackListModel):

    changed = pyqtSignal(int, int)

    def __init__(self, app, collection):
        TrackListModel.__init__(self, app, collection)
        return

    def _selection_check_remove_index(self, sel, i):
        if i in sel:
            # remove stale indices from
            del sel[sel.index(i)]
            # leave selection in reasonable state
            if not sel:
                if len(self._data) > 0:
                    sel.append(0)
                else:
                    sel.append(-1)
        return

    def do_get_document(self, index):
        docid = self[index].docid
        return self.collection.db.get_document(docid), docid

    def do_get_docid(self, index):
        if 0 <= index <= len(self._data)-1:
            return self[index].docid
        return -1

    def do_get_index_from_docid(self, targetid):
        return self.index_docid(targetid)

    def unqueue_docid(self, docid):
        i = self.index_docid(docid)

        self.beginRemoveRows(QModelIndex(), i, i)
        doc = self._data.pop(i)
        self._count -= 1
        self.endRemoveRows()

        self.changed.emit(doc.docid, self.QUEUE_EVENT_REMOVED)
        self.currentStreamIndexChanged.emit()
        return

    def do_unqueue_selection(self):
        for i in self.selected_indices:
            self.unqueue_index(i)
        return

    def unqueue_index(self, i):
        if i == -1: return
        self.beginRemoveRows(QModelIndex(), i, i)
        doc = self._data.pop(i)
        self._count -= 1
        self.endRemoveRows()

        self.changed.emit(doc.docid, self.QUEUE_EVENT_REMOVED)
        self.currentStreamIndexChanged.emit()
        return

    def enqueue_docid(self, docid):
        i = self._count

        self.beginInsertRows(QModelIndex(), i, i)
        t = Track.from_docid(self, docid)
        self.append(t)
        self._count += 1
        self.endInsertRows()

        self.changed.emit(docid, self.QUEUE_EVENT_ADDED)
        self.currentStreamIndexChanged.emit()
        return

    @pyqtSlot()
    def empty(self):
        docids = [track.docid for track in self._data]

        self.beginResetModel()
        Queue.clear(self)
        self._count = 0
        self.endResetModel()

        for docid in docids:
            self.changed.emit(docid, self.QUEUE_EVENT_REMOVED)
        self.currentStreamIndexChanged.emit()
        return

    @pyqtProperty(int, notify=changed)
    def count(self):
        return self._count

    def cleanup_historic(self):
        sel = self.selected_indices
        for i, track in enumerate(self._data):
            if track.historic:
                # update list content
                self.beginRemoveRows(QModelIndex(), i, i)
                del self._data[i]
                self._count -= 1
                self.endRemoveRows()
                self.changed.emit(track.docid, self.QUEUE_EVENT_REMOVED)
                # update current selection & stream index
                self._selection_check_remove_index(sel, i)
                self.currentStreamIndexChanged.emit()
        return

    def do_get_current_stream_index(self):
        return self._current_stream_index

    def do_data(self, index, role, roles):
        if not index.isValid() or not self._count:
            return None

        doc, docid = self.get_document(index.row())
        rolev = self.ROLES[role]

        if rolev == self.ROLE_TRACKNUM:
            return doc.get_value(DbEnums.V_TRACKNUM).decode("utf8", "ignore")
        elif rolev == self.ROLE_TITLE:
            return doc.get_value(DbEnums.V_TITLE).decode("utf8", "ignore")
        elif rolev == self.ROLE_ALBUM:
            return doc.get_value(DbEnums.V_ALBUM).decode("utf8", "ignore")
        elif rolev == self.ROLE_ARTIST:
            return doc.get_value(DbEnums.V_ARTIST).decode("utf8", "ignore")
        elif rolev == self.ROLE_DURATION:
            return doc.get_value(DbEnums.V_DURATION_FSTR)
        elif rolev == self.ROLE_TAGS:
            return doc.get_value(DbEnums.V_TAGS).decode("utf8", "ignore")
        elif rolev == self.ROLE_ENQUEUED:
            return True # it's in the queue, we can assume its enqueued ;)
        elif rolev == self.ROLE_LIKED:
            return doc.get_value(DbEnums.V_LIKE) or '0'
        elif rolev == self.ROLE_SELECTED:
            return index.row() in self.selected_indices
        return None

    def update(self):
        self.cleanup_historic()
        TrackListModel.update(self)
        return

    def next_track(self):
        for track in self._data:
            if not track.historic:
                track.historic = True
                return track
        return None

    @pyqtSlot(float, result=int)
    def shift_selected(self, delta, parent=QModelIndex()):
        if abs(delta) < 1.0:
            return -1

        # we allow only one row to be shifted
        from_ = self.selected_indices[0]
        to = min(self._count, max(0, from_ + int(delta - 0.5)))

        self.beginRemoveRows(parent, from_, from_)
        t = self._data.pop(from_)
        self.endRemoveRows()

        self.beginInsertRows(parent, to, to)
        self._data.insert(to, t)
        self.endInsertRows()

        self.update_selection(to)

        self.currentStreamIndexChanged.emit()
        return to
