#
# Authors:
#  Michael Vogt
#  Ollivier Tilloy
#  Matthew McGowan
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import json
import threading
import os.path
import urllib
import urlparse

from PyQt4.QtCore import QAbstractListModel, QModelIndex, pyqtSlot

from totallywired.utils.caching import Caches


def get_station_favico(host):
    favico_path = os.path.join(Caches.RADIO, host)
    if not os.path.exists(favico_path):
        quoted_url = urllib.quote("http://"+host)
        favico = urllib.urlopen(_RadioStationMetadataWorker.GOOGLE_FAVICON_FINDER % quoted_url)
        with open(favico_path, "wb") as f:
            f.write(favico.read())
    return favico_path


class _FavicoLoader(threading.Thread):

    def __init__(self, uri, finished_cb, i):
        threading.Thread.__init__(self)
        self.host = urlparse.urlparse(uri).hostname
        self.finished_cb = finished_cb
        self.i = i
        return

    def run(self):
        get_station_favico(self.host)
        self.finished_cb(self.i)
        return


class _RadioStationMetadataWorker(threading.Thread):

    GOOGLE_FAVICON_FINDER = "http://www.google.com/s2/favicons?domain_url=%s"

    def __init__(self, uri, internal_cb):
        threading.Thread.__init__(self)
        self.daemon = True
        self.uri = uri
        self.icb = internal_cb
        return

    def run(self):
        import lxml.html
        title = ""
        host = urlparse.urlparse(self.uri).hostname
        try:
            t = lxml.html.parse("http://"+host)
            title = t.find(".//title").text
        except Exception as e:
            print(e)

        if not title:
            try:
                t = lxml.html.parse(self.uri)
                title = t.find(".//title").text
            except Exception as e:
                print(e)
                return

        favico_path = get_station_favico(host)
        self.icb(title, favico_path)
        return


class RadioStationModel(QAbstractListModel):

    ROLE_SOURCE = "_source"
    ROLE_STATE = "_state"
    ROLE_SRC_ICON = "_srcicon"
    ROLE_SRC_NAME = "_srcname"
    ROLE_DISPL_TEXT = "_displ_text"
    ROLE_IS_PLAYING = "_is_playing"

    COLUMNS = (ROLE_SOURCE, ROLE_STATE,
               ROLE_SRC_ICON, ROLE_SRC_NAME, ROLE_DISPL_TEXT, ROLE_IS_PLAYING)

    ROLES = dict(enumerate(COLUMNS))

    def __init__(self):
        QAbstractListModel.__init__(self)
        self.setRoleNames(RadioStationModel.ROLES)
        self._count = 0
        self._cur = -1
        self.player = None
        return

    def state_change(self, *args):
        i = self._cur
        self.dataChanged.emit(
            self.createIndex(i, 0),
            self.createIndex(i, 0))
        return

    def update_stations(self):
        self.clear()
        path = os.path.join(Caches.RADIO, "radio.json")
        with open(path, "r") as f:
            try:
                self._data = json.load(f)
            except:
                self._data = []
        self._count = len(self._data)
        self.beginInsertRows(QModelIndex(), 0, self._count-1)
        self.endInsertRows()
        return

    @pyqtSlot(str)
    def add_station(self, uri):
        def on_station_metadata_ready(title, favicon):
            _json = open(os.path.join(Caches.RADIO, "radio.json"))
            try:
                d = json.load(_json)
            except:
                d = []
            _json.close()

            station = {"uri":str(uri), "name":title, "favicon":favicon}
            d.append(station)

            s = json.dumps(d)
            _json = open(os.path.join(Caches.RADIO, "radio.json"), "w")
            _json.write(s)
            _json.close()

            self.update_stations()

        worker = _RadioStationMetadataWorker(str(uri), on_station_metadata_ready)
        worker.start()
        return

    def _download_favicon(self, i):
        def on_finished(i):
            return

        job = _FavicoLoader(self._data[i]["uri"], on_finished, i)
        job.start()
        return

    def _get_index(self, src):
        for i, s in enumerate(self._data):
            if s["uri"] == src:
                return i
        return -1

    def clear(self):
        self.beginResetModel()
        self._count = 0
        self.endResetModel()
        return

    # QAbstractListModel compliance code
    def rowCount(self, parent=QModelIndex()):
        return self._count

    def data(self, index, role, roles=ROLES):
        if not index.isValid():
            return None

        rolev = self.ROLES[role]
        station = self._data[index.row()]

        if rolev == self.ROLE_SOURCE:
            return station["uri"]
        elif rolev == self.ROLE_SRC_NAME:
            return station["name"]
        elif rolev == self.ROLE_SRC_ICON:
            path = os.path.join(Caches.RADIO, station["favicon"])
            if not os.path.exists(path):
                self._download_favicon(index.row())
            return path
        elif rolev == self.ROLE_DISPL_TEXT:
            return ""
        elif rolev == self.ROLE_IS_PLAYING:
            if self.player is None:
                return False
            elif self.player.track is None:
                return False
            elif not self.player.is_playing:
                return False
            return station["uri"] == self.player.track.uri
        return None

    @pyqtSlot(int)
    def stop(self, i):
        self.player.stop()
        self._cur = -1
        self.dataChanged.emit(
            self.createIndex(i, 0),
            self.createIndex(i, 0))
        return

    @pyqtSlot(int)
    def remove_station(self, i):
        if i == self._cur:
            self.player.stop()
            self._cur = -1

        self.beginRemoveRows(QModelIndex(), i, i)
        del self._data[i]
        self._count -= 1
        self.endRemoveRows()

        _json = open("data/radio/radio.json", "w")
        s = json.dumps(self._data)
        _json.write(s)
        _json.close()
        return

    @pyqtSlot(int)
    def start(self, i):
        pi = self._cur   # previous index
        station = self._data[i]

        self.player.start_radio(station)

        self._cur = i
        self.dataChanged.emit(
            self.createIndex(i, 0),
            self.createIndex(i, 0))

        if self._cur == -1:
            return

        self.dataChanged.emit(
            self.createIndex(pi, 0),
            self.createIndex(pi, 0))
        return
