import os
import dbus
import time
import string
import traceback

from gi.repository import GLib
from PyQt4.QtCore import (
    QObject, pyqtSlot, pyqtProperty, pyqtSignal)

from totallywired.db.collectors.enums import DbEnums

from totallywired.player.common import (
    SECOND,
    State,
    Track,
    History,
    ContentPending,
    RadioStationLiveStream)

from totallywired.utils.misc import (
    StdoutColors,
    get_nice_hms,
    tail)


class QMusicPlayerClient(QObject):

    SERVICE_COMM_ATTEMPT_COUNT = 20
    SERVICE_COMM_ATTEMPT_DELAY = 0.1

    playbackStateChanged = pyqtSignal()
    selectionModeChanged = pyqtSignal()
    currentStreamChanged = pyqtSignal()
    playbackPositionChanged = pyqtSignal()
    mediaKeyPressed = pyqtSignal(int)

    MEDIAKEY_PLAY = 'Play'
    MEDIAKEY_NEXT = 'Next'
    MEDIAKEY_PREVIOUS = 'Previous'

    TICK_TIME = 500

    _ID = 'tw-player-client'

    def __init__(self, app, music, queue, bus):
        QObject.__init__(self)

        self.app = app
        self.music = music
        self.track = None

        settings = app.settings
        settings.beginGroup("musicplayer")

        self.repeat = settings.value("repeat").toBool()
        self.shuffle = settings.value("shuffle").toBool()

        settings.endGroup()

        self.history = History()
        self.queue = queue

        self._syncd = False
        self._state = [State.NULL.value_nick,]
        self._duration = -1
        self._position = -1
        self._seeking = False

        attempt = 0
        self.player = None
        while attempt <= self.SERVICE_COMM_ATTEMPT_COUNT:
            try:
                self.player = bus.get_object('org.tw.MusicPlayerService', '/MusicPlayerService')
                break
            except dbus.DBusException:
                attempt += 1
                print(StdoutColors.BOLD + StdoutColors.WARNING + "No service 'org.tw.MusicPlayerService', retrying... (attempt %s/%s)" % (attempt, self.SERVICE_COMM_ATTEMPT_COUNT) + StdoutColors.ENDC)
                time.sleep(self.SERVICE_COMM_ATTEMPT_DELAY)

        if self.player is not None:
            bus.add_signal_receiver(
                self.on_playback_error_signal,
                dbus_interface='org.tw.MusicPlayerService',
                signal_name='PlaybackError')
            bus.add_signal_receiver(
                self.on_playback_eos_signal,
                dbus_interface='org.tw.MusicPlayerService',
                signal_name='PlaybackEOS')
            bus.add_signal_receiver(
                self.on_audio_changed_signal,
                dbus_interface='org.tw.MusicPlayerService',
                signal_name='AudioChanged')
        else:
            print("Service 'org.tw.MusicPlayerService' could not be contacted!!")

        try:
            self.mediakeys = bus.get_object(
                'org.gnome.SettingsDaemon',
                '/org/gnome/SettingsDaemon/MediaKeys')
        except dbus.DBusException:
            self.mediakeys = None
            traceback.print_exc()

        # this is what gives us the multi media keys.
        self.grab_mediakeys()

        # connect_to_signal registers our callback function.
        self.mediakeys.connect_to_signal(
            'MediaPlayerKeyPressed', self._handle_mediakey)

        # player refs for the music and queue models
        self._link_models(music, music.cdda, queue, app.stations)
        # timer updates playback position
        self.startTimer(self.TICK_TIME)
        return

    # dbus signal handlers
    def _handle_mediakey(self, app, event):
        event = str(event)
        print('MediaKeyEvent:', event)
        if str(app) != self._ID:
            return
        if event == self.MEDIAKEY_PLAY:
            if self.is_playing:
                self.pause()
                self.mediaKeyPressed.emit(0)
            elif self.is_paused:
                self.resume()
                self.mediaKeyPressed.emit(1)
            else:
                self.mediaKeyPressed.emit(1)
                next(self)
        elif event == self.MEDIAKEY_NEXT:
                self.mediaKeyPressed.emit(1)
                next(self)
        elif event == self.MEDIAKEY_PREVIOUS:
                self.mediaKeyPressed.emit(1)
                self.previous()

    def on_playback_error_signal(self):
        self.stop()
        return

    # End of Stream (EOS)
    def on_playback_eos_signal(self):
        next(self)
        return

    def on_audio_changed_signal(self, uri):
        uri = str(uri)
        print(uri)
        if (uri and not uri.startswith('file:')) or not uri:
            self.track.url = uri
            print("Streaming from: %s" % uri)
        else:
            self.track = Track.from_uri(self.music, uri)

        self._position = 0
        self._duration = 0

        self.playbackStateChanged.emit()
        self.currentStreamChanged.emit()
        return

    # QObject virtual functions
    def timerEvent(self, event):
        self._tick()
        return

    # private
    def _link_models(self, *models):
        def on_stream_change(*args):
            for m in models:
                if hasattr(m, "update"):
                    m.update()

        def on_state_change(*args):
            for m in models:
                if hasattr(m, "state_change"):
                    m.state_change(*args)

        def on_queue_change(*args):
            for m in models:
                if hasattr(m, "queue_changed"):
                    m.queue_changed(*args)

        for m in models:
            m.player = self

        self.queue.changed.connect(on_queue_change)
        self.currentStreamChanged.connect(on_stream_change)
        self.playbackStateChanged.connect(on_state_change)
        return

    def _tick(self):
        if State.PLAYING.value_nick in self.player.get_state():
            if self._duration <= 0:
                pos, dur = self.player.get_positional_state()
                self._position = int(pos)
                self._duration = int(dur)
            else:
                self._position += self.TICK_TIME / 1000.0 * SECOND
            self.playbackPositionChanged.emit()
        return

    def sync_with_remote_player(self):
        uri = str(self.player.get_current_uri())
        path = False

        try:
            path = GLib.filename_from_uri(uri, "")
        except:
            pass

        if path:
            if path and os.path.exists(path):
                pos, dur = self.player.get_positional_state()
                self._position = float(pos)
                self._duration = float(dur)
                self._state = [str(s) for s in self.player.get_state()]
                self.track = Track.from_uri(self.music, uri)
                self.music.update()
                print("Syncd: %s %s %s/%s (ns)" % (tail(uri), self._state, self._position, self._duration))
        else:
            self.stop()

        self.playbackStateChanged.emit()
        self.currentStreamChanged.emit()
        self.playbackPositionChanged.emit()
        self._syncd = True
        return

    def _start_track(self, track, index=None):
        if self.app.ignored:
            return True

        self._duration = 0
        self._position = 0
        res = self.player.start(track.uri)
        if bool(res):
            self.track = track
            self._state = [State.PLAYING.value_nick,]
            self.playbackStateChanged.emit()
            self.currentStreamChanged.emit()
        else:
            self.stop()
        return True

    # public
    def grab_mediakeys(self):
        if self.player is None or self.mediakeys is None:
            return

        self.mediakeys.GrabMediaPlayerKeys(
            self._ID, 0,
            dbus_interface='org.gnome.SettingsDaemon.MediaKeys')
        return

    def get_current_title(self):
        if self.track is None or self.is_stopped:
            return ""
        return self.track.title(self.music)

    def get_current_album_info(self):
        if self.track is None or self.is_stopped:
            return ""
        elif isinstance(self.track, RadioStationLiveStream):
            return "<i>streaming from</i> %s" % self.track.url or self.track.uri
        doc = self.music.get_document_from_docid(self.track.docid)
        album = str(doc.get_value(DbEnums.V_ALBUM)).decode("utf-8", "ignore")
        artist = str(doc.get_value(DbEnums.V_ARTIST)) or "Unknown"
        date = str(doc.get_value(DbEnums.V_DATE)).decode("utf-8", "ignore")
        s = []

        if album and date:
            s.append('<i>from</i>  <b>%s (%s)</b>  ' % (album, date))
        elif album:
            s.append('<i>from</i>  <b>%s</b>  ' % album)
        s.append('<i>by</i>  <b>%s</b>' % artist)
        if not album and date:
            s.append('<b>  (%s)</b>' % date)
        return string.join(s)

    def update_docids(self):
        if self.music is None:
            return
        db = self.music._db
        if self.track is not None:
            ret = self.track.update_docid(db)
            if not ret:
                print("Current track no longer available!, Stopping playback...")
                self.stop()

        self.queue.update_docids(db)
        self.history.update_docids(db)
        return

    @pyqtSlot(str)
    def start_radio(self, station):
        if self.app.ignored:
            return True

        def error_handler(*error):
            return

        def reply_handler(res):
            if bool(res):
                self.track = RadioStationLiveStream(station)
                self._state = [State.PLAYING.value_nick,]
                self.playbackStateChanged.emit()
                self.currentStreamChanged.emit()
            else:
                self.stop()
            return

        self._duration = 0
        self._position = 0
        self.track = ContentPending()
        self.playbackStateChanged.emit()

        self.player.start(
            station["uri"],
            reply_handler=reply_handler,
            error_handler=error_handler)
        return True

    @pyqtProperty(bool, notify=playbackPositionChanged)
    def duration_known(self):
        return self._duration > 0

    @pyqtProperty(bool, notify=playbackStateChanged)
    def is_buffering_or_streaming(self):
        return isinstance(self.track, (ContentPending, RadioStationLiveStream))

    @pyqtProperty(str, notify=playbackStateChanged)
    def current_track_title(self):
        return self.get_current_title()

    @pyqtProperty(str, notify=playbackStateChanged)
    def current_album_info(self):
        return self.get_current_album_info()

    @pyqtProperty(bool, notify=playbackStateChanged)
    def is_playing(self):
        if self.player is None:
            return False
        return (State.PLAYING.value_nick in self._state)

    @pyqtProperty(bool, notify=playbackStateChanged)
    def is_paused(self):
        return (State.PAUSED.value_nick in self._state)

    @pyqtProperty(bool, notify=playbackStateChanged)
    def is_stopped(self):
        if self.player is None:
            return True
        return (State.NULL.value_nick in self._state)

    @pyqtProperty(float, notify=playbackPositionChanged)
    def playback_position(self):
        if self._duration <= 0.0:
            return 0.0
        return min(1.0, float(self._position) / self._duration)

    @pyqtProperty(str, notify=playbackPositionChanged)
    def nice_playback_position_info(self):
        if self._duration <= 0:
            return u'%s <font color="#525252">/ \u221E</font>' % get_nice_hms(self._position/SECOND)
        return '%s <font color="#525252">/ %s</font>' % (get_nice_hms(self._position/SECOND), get_nice_hms(self._duration/SECOND))

    @pyqtSlot(float, result=str)
    def position_to_time(self, position):
        return "%s" % get_nice_hms(position * self._duration/SECOND)

    @pyqtSlot(float)
    def set_position(self, position):
        if self.app.ignored:
            return True
        self._position = int(self._duration * position)
        self.player.seek_to(str(self._position))
        self.playbackPositionChanged.emit()
        return

    @pyqtSlot(result=bool)
    def stop(self):
        if self.app.ignored:
            return True
        self.track = None
        self._state = [State.NULL.value_nick,]
        res = self.player.stop()
        self.playbackStateChanged.emit()
        self.music.update()
        return res

    @pyqtSlot(result=bool)
    def pause(self):
        if self.app.ignored:
            return True

        self._state = [State.PAUSED.value_nick,]
        res = self.player.pause()
        self.playbackStateChanged.emit()
        self.music.update()
        return res

    @pyqtSlot(result=bool)
    def resume(self):
        if self.app.ignored:
            return True

        self._state = [State.PLAYING.value_nick,]
        res = self.player.resume()
        self.playbackStateChanged.emit()
        self.music.update()
        return res

    @pyqtSlot(result=bool)
    def previous(self):
        if self.app.ignored:
            return True

        track = self.history.step_back()
        if track is not None:
            res = self._start_track(track)
        else:
            res = False
        return res

    @pyqtSlot(result=bool)
    def next(self):
        if self.app.ignored:
            return True

        self.history.reset()
        if self.track is not None and not self.track.historic:
            self.history.append(self.track)

        docid, index = self.music.get_next()
        if docid:
            track = Track.from_docid(self.music, docid)
            return self._start_track(track, index)

        self.stop()
        return False

    @pyqtProperty(bool, notify=currentStreamChanged)
    def has_history(self):
        return (self.is_playing or self.is_paused) and not self.history.at_begining()

    @pyqtProperty(bool, notify=playbackStateChanged)
    def has_stream(self):
        return (not self.is_stopped)

    @pyqtProperty(bool, notify=selectionModeChanged)
    def shuffle_active(self):
        return self.shuffle

    @pyqtProperty(bool, notify=selectionModeChanged)
    def repeat_active(self):
        return self.repeat

    @pyqtSlot(bool)
    def set_shuffle(self, shuffle):
        if self.app.ignored:
            return True

        self.shuffle = shuffle
        self.selectionModeChanged.emit()
        return

    @pyqtSlot(bool)
    def set_repeat(self, repeat):
        if self.app.ignored:
            return True

        self.repeat = repeat
        self.selectionModeChanged.emit()
        return
