import gst as Gst

from totallywired.db.collectors.enums import DbEnums
from totallywired.utils.db import get_qterm
from totallywired.utils.misc import tail


SECOND = Gst.SECOND


def _get_docid(db, uri):
    if db is None:
        return -1
    pl = db.postlist(get_qterm(uri))
    try:
        plitem = next(pl)
    except StopIteration:
        return -1
    return plitem.docid


class State:
    #~ NULL = State.NULL
    #~ PLAYING = State.PLAYING
    #~ PAUSED = State.PAUSED

    NULL = Gst.STATE_NULL
    PLAYING = Gst.STATE_PLAYING
    PAUSED = Gst.STATE_PAUSED


class Queue(object):

    def __init__(self):
        self._data = list()
        return

    def __data__(self):
        return [t for t in self._data if not t.historic]

    def __len__(self):
        return len(self._data)

    def __delitem__(self, index):
        return self._data.__delitem__(index)

    def __setitem__(self, index, v):
        return self._data.__setitem__(index, v)

    def __getitem__(self, index):
        return self._data.__getitem__(index)

    def append(self, track):
        if not isinstance(track, Track):
            print("Error: got %s not Track" % track)
            return False, -1
        elif self._data and track.docid == self._data[-1].docid:
            print('Skipping because consecutive docids matched!')
            return False, -1
        self._data.append(track)
        return True, len(self._data)

    def extend(self, many_tracks):
        self._data.extend(many_tracks)
        return True, len(self._data)

    def remove_by_docid(self, docid):
        if docid in self._data:
            i = self._data.index(docid)
            del self._data[i]
            return True, i
        return False, -1

    def has_docid(self, docid):
        return docid in self.__data__()

    def index_docid(self, docid):
        if self.has_docid(docid):
            return self._data.index(docid)
        return -1

    def has_uri(self, uri):
        return uri in self._data

    def has_track(self, track):
        return track in self._data

    def has_tracks(self):
        return len([t for t in self._data if not t.historic]) > 0

    def update_docids(self, db):
        for track in self._data:
            track.update_docid(db)
        return

    def is_last(self, track):
        if len(self._data) == 0:
            return False
        return self._data[-1] == track

    def clear(self):
        self._data = []
        return


class History(Queue):

    def __init__(self):
        Queue.__init__(self)
        self._cursor = 0
        return

    def at_begining(self):
        return self._cursor < 1

    def step_back(self):
        if self._cursor <= 0:
            print("At the begining of time!!!")
            return None
        self._cursor -= 1
        print('History:step back', self._cursor)
        return self._data[self._cursor]

    def reset(self):
        self._cursor = len(self._data)
        return

    def append(self, track):
        track.historic = True
        res = Queue.append(self, track)
        self._cursor = len(self._data)
        if res:
            print('History:append', self._cursor)
        return


class Track(object):

    @staticmethod
    def from_docid(model, docid):
        if docid == -1:
            uri = ""
        else:
            uri = model.get_document_from_docid(docid).get_data()
        return Track(uri, docid)

    @staticmethod
    def from_uri(model, uri):
        docid = model.get_docid_from_uri(uri)
        return Track(uri, docid)

    def __init__(self, uri, docid):
        self.uri = uri
        self.docid = int(docid)
        self.active = True
        self.historic = False
        return

    def __eq__(self, other):
        if isinstance(other, Track):
            return other.docid == self.docid
        elif isinstance(other, (long, int)):
            return int(other) == self.docid
        elif isinstance(other, str):
            return other == self.uri
        return False

    def __str__(self):
        return "Track(.../%s:%s)" % (tail(self.uri), self.docid)

    def __repr__(self):
        return str(self)

    def update_docid(self, db):
        self.active = False
        self.docid = _get_docid(db, self.uri)
        if self.docid > 0:
            self.active = True
        else:
            print(("Could not update docid for doc with uri: %s" % self.uri))
        return self.active

    def title(self, model):
        doc = model.get_document_from_docid(self.docid)
        return "<b>%s</b>" % str(doc.get_value(DbEnums.V_TITLE)).decode("utf-8", "ignore")


class RadioStationLiveStream(Track):

    TITLE = 0

    @staticmethod
    def from_docid(model, docid):
        raise NotImplementedError

    @staticmethod
    def from_uri(model, uri):
        raise NotImplementedError

    def __init__(self, station):
        Track.__init__(self, station["uri"], 0)
        self.url = ""
        self._title = station["name"]
        return

    def update_docid(self, db):
        return True

    def title(self, *ignored):
        return "<b>%s</b>" % (self._title or "Live Radio")


class ContentPending(Track):

    @staticmethod
    def from_docid(model, docid):
        raise NotImplementedError

    @staticmethod
    def from_uri(model, uri):
        raise NotImplementedError

    def __init__(self):
        Track.__init__(self, "", 0)
        return

    def update_docid(self, db):
        return True

    def title(self, *ignored):
        return "Content starting..."
