import os
import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

#~ from gi.repository import GObject, Gst
#~ Gst.init(None)

from gi.repository import TotemPlParser

import gobject as GObject
GObject.threads_init()
from gi.repository import GLib
import pygst
pygst.require("0.10")
import gst as Gst

from .common import State
from totallywired.utils.misc import ellipsis
from totallywired.utils.paths import is_cdda_gvfs, get_cdda_gvfs


def cdda_uri_from_uri(uri):
    track = os.path.split(uri)[-1]
    cdda = get_cdda_gvfs()
    return os.path.join("file://" + cdda, track)

playlist_exts = ('m3u', 'asx', 'pls')
def is_playlist(uri):
    for ext in playlist_exts:
        if uri.endswith(ext):
            return True
    return False


class MusicPlayerService(dbus.service.Object):

    def __init__(self, mainloop):
        session_bus = dbus.SessionBus()
        bus_name = dbus.service.BusName('org.tw.MusicPlayerService', bus=session_bus)
        dbus.service.Object.__init__(self, bus_name, '/MusicPlayerService')

        self.mainloop = mainloop

        #~ self.player = Gst.ElementFactory.make("playbin2", "player")
        #~ fakesink = Gst.ElementFactory.make("fakesink", "fakesink")

        self.player = Gst.element_factory_make("playbin2", "player")
        fakesink = Gst.element_factory_make("fakesink", "fakesink")
        self.player.set_property("video-sink", fakesink)
        self.player.connect("audio-changed", self.on_audio_changed)

        bus = self.player.get_bus()
        bus.add_signal_watch()
        bus.connect("message::eos", self.on_eos)
        bus.connect("message::error", self.on_error)

        self._job = 0
        return

    def on_eos(self, bus, message):
        self._set_state(State.NULL)
        self.PlaybackEOS()
        return

    def on_error(self, bus, message):
        err, debug = message.parse_error()
        print("%s.Error: %s, %s" % (self._job, err, debug))
        self.PlaybackError()
        return

    def on_audio_changed(self, *args):
        uri = self.player.get_property("uri")
        self.AudioChanged(uri)
        print("%s.NowPlaying: %s" % (self._job, ellipsis(uri, maxlen=50)))
        return

    def _set_state(self, state):
        return self.player.set_state(state)

    def _get_state(self):
        return [s.value_nick for s in self.player.get_state()]

    def _handle_playlist(self, uri, entries):

        def on_entry_parsed(parser, uri, metadata):
            entries.append(uri)
            return

        parser = TotemPlParser.Parser()
        parser.connect("entry-parsed", on_entry_parsed)

        parser.parse(uri, False)
        if entries:
            self.player.set_property("uri", entries[0])
            self._set_state(State.PLAYING)
            return True
        return False

    @dbus.service.signal('org.tw.MusicPlayerService')
    def PlaybackError(self):
        pass

    @dbus.service.signal('org.tw.MusicPlayerService')
    def PlaybackEOS(self):
        pass

    @dbus.service.signal('org.tw.MusicPlayerService')
    def AudioChanged(self, uri):
        pass

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='s', out_signature='b')
    def seek_to(self, ns):
        event = Gst.event_new_seek(
            1.0, Gst.FORMAT_TIME,
            Gst.SEEK_FLAG_FLUSH | Gst.SEEK_FLAG_ACCURATE,
            Gst.SEEK_TYPE_SET,
            int(ns),
            Gst.SEEK_TYPE_NONE, 0)

        res = self.player.send_event(event)

        if res:
            self.player.set_new_stream_time(0)
        print('%s.Seeking to %s ns...' % (self._job, ns))
        return res

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='s', out_signature='b')
    def start(self, uri):
        uri = str(uri)
        self._set_state(State.NULL)
        self._job += 1
        if is_cdda_gvfs(uri):
            uri = cdda_uri_from_uri(uri)
        elif is_playlist(uri):
            return self._handle_playlist(uri, [])
        self.player.set_property("uri", uri)
        self._set_state(State.PLAYING)
        return True

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='', out_signature='s')
    def get_current_uri(self):
        uri = self.player.get_property("uri")
        if not uri:
            uri = ""
        return uri

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='', out_signature='as')
    def get_state(self):
        return self._get_state()

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='', out_signature='b')
    def resume(self):
        self._set_state(State.PLAYING)
        print("%s.Resumed" % self._job)
        return True

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='', out_signature='b')
    def pause(self):
        self._set_state(State.PAUSED)
        print("%s.Paused" % self._job)
        return True

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='', out_signature='b')
    def stop(self):
        self._set_state(State.NULL)
        print("%s.Stopped!" % self._job)
        return True

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='', out_signature='(dd)')
    def get_positional_state(self):
        try:
            pos = self.player.query_position(Gst.FORMAT_TIME, None)[0]
        except:
            pos = -1
        try:
            dur = self.player.query_duration(Gst.FORMAT_TIME, None)[0]
        except:
            dur = -1
        return pos, dur

    @dbus.service.method('org.tw.MusicPlayerService', in_signature='', out_signature='')
    def quit(self):
        self.mainloop.quit()
        return
