import QtQuick 1.1
import QtDesktop 0.1

Item {
    ListView {
        id: indexes
        model: Indexes

        clip: true
        boundsBehavior: Flickable.StopAtBounds

        anchors.left: parent.left
        anchors.right: parent.right
        anchors.top: sourceHeader.bottom
        anchors.bottom: parent.bottom

        spacing: 6

        header: Item {
            height: 12
            width: 1
        }

        section.property: "_srctype"
        section.criteria: ViewSection.FullString
        section.delegate: Item {
            height: text.paintedHeight + 10
            anchors.horizontalCenter: parent.horizontalCenter
            width: Math.min(800, parent.width)

            TextLabel {
                id: text
                text: section
                font.bold: true
                font.pixelSize: 20
                color: "#919090"
                style: Text.Raised
                styleColor: "#FFFDFD"
                x: 0; y: 5
            }
        }

        delegate: Rectangle {
            id: row

            property string source: _source
            property string indexsrc: _index
            property string istate: _state
            property string progress: _progress
            property string srcicon: _srcicon
            property string srcname: _srcname
            property int doccount: _doccount
            property string displ_text: _displ_text

            radius: 5
            clip: true

            color: {
                if (Indexes.state_is(istate, "not-connected"))
                    return "#BFB6AC"
                else
                    if (Indexes.state_is(istate, "hidden"))
                        return "#6D2E56"
                return "#F79658"
            }

            Behavior on color { PropertyAnimation {duration: 300} }

            anchors.horizontalCenter: parent.horizontalCenter
            height: 90
            width: Math.min(800, parent.width)

            Text {
                font.pixelSize: 32
                font.letterSpacing: -3
                font.bold: true
                anchors.margins: 12
                color: Qt.lighter(row.color, 1.2)
                style: Text.Raised
                styleColor: Qt.darker(row.color, 1.225)
                anchors.verticalCenter: parent.verticalCenter
                anchors.verticalCenterOffset: 18
                anchors.right: parent.right
                text: displ_text
            }

            Image {
                id: icon
                width: 48; height: 48;
                fillMode: Image.PreserveAspectFit
                smooth: true
                source: srcicon
                anchors.top: parent.top
                anchors.left: parent.left
                anchors.leftMargin: 12
                anchors.topMargin: 12
            }

            Column {
                anchors.verticalCenter: parent.verticalCenter
                anchors.left: icon.right
                anchors.right: parent.right

                anchors.margins: 12
                spacing: 6

                Text {
                    // name
                    anchors.left: parent.left
                    anchors.right: parent.right
                    color: Qt.darker(row.color, 2.2)
                    font.bold: true
                    font.pointSize: 12
                    text: srcname
                    style: Text.Raised
                    styleColor: Qt.lighter(row.color, 1.225)
                }

                Text {
                    // source path
                    anchors.left: parent.left
                    anchors.right: parent.right
                    color: Qt.darker(row.color, 2.2)
                    text: "<small>" + source + "</small>"
                    style: Text.Raised
                    styleColor: Qt.lighter(row.color, 1.225)
                }

                Text {
                    // audiofile contribution
                    anchors.left: parent.left
                    anchors.right: parent.right
                    color: Qt.darker(row.color, 2.2)
                    text: "<small>This source contributes <b>" + doccount + "</b> audiofiles to your music collection</small>"
                    style: Text.Raised
                    styleColor: Qt.lighter(row.color, 1.225)
                }
            }

            Row {
                anchors.margins: 12
                anchors.right: parent.right
                anchors.top: parent.top

                opacity: progress > -1 ? 0 : 1
                Behavior on opacity { PropertyAnimation {duration: 300} }

                ListViewButton {
                    opacity: Indexes.state_is(istate, "ready") ? 1 : 0

                    label: "g"
                    font.family: "TotallyWired"
                    font.pixelSize: 20
                    verticalPadding: 4
                    horizontalPadding: 4
                    baseColor: row.color

                    onClicked: {
                        if (Indexes.state_is(istate, "ready")) {
                            DbClient.build_collection(source)
                        }
                    }
                }

                ListViewButton {
                    opacity: Indexes.state_is(istate, "ready|hidden") ? 1 : 0

                    label: "e"
                    font.family: "TotallyWired"
                    font.pixelSize: 20
                    verticalPadding: 4
                    horizontalPadding: 4
                    baseColor: row.color

                    onClicked: {
                        if (!Indexes.state_is(istate, "indexing")) {
                            DbClient.toggle_hidden(source)
                        }
                    }
                }

                ListViewButton {
                    opacity: (
                        Indexes.state_is(istate, "ready|not-connected|hidden") &
                        !Indexes.state_is(istate, "cdda") ?
                         1 : 0)

                    label: "f"
                    font.family: "TotallyWired"
                    font.pixelSize: 20
                    verticalPadding: 4
                    horizontalPadding: 4
                    baseColor: row.color

                    onClicked: {
                        okLoader.source = "InlineOkDialog.qml"
                    }
                }
            }

            ListView.onAdd: SequentialAnimation {
                PropertyAction { target: row; property: "height"; value: 0 }
                NumberAnimation { target: row; property: "height"; to: 100; duration: 150; easing.type: Easing.InOutQuad }
            }

            ListView.onRemove: SequentialAnimation {
                PropertyAction { target: row; property: "ListView.delayRemove"; value: true }
                NumberAnimation { target: row; property: "height"; to: 0; duration: 150; easing.type: Easing.InOutQuad }
                // Make sure delayRemove is set back to false so that the item can be destroyed
                PropertyAction { target: row; property: "ListView.delayRemove"; value: false }
            }

            Loader {
                id: okLoader
                anchors.fill: parent
            }

            Connections {
                target: okLoader.item
                onYesClicked: {
                    okLoader.source = ""
                    DbClient.trash_collection(source)
                    Indexes.remove_index(index)
                }
                onNoClicked: {
                    okLoader.source = ""
                }
            }
        }

        footer: Item {
            height: 48
        }
    }

    Item {
        id: sourceHeader
        height: addSource.y + addSource.height + 24 + 36
        anchors.horizontalCenter: parent.horizontalCenter
        width: Math.min(800, parent.width)

        Column {
            spacing: 12

            anchors.top: parent.top
            width: parent.width

            anchors.topMargin: 36

            Button {
                id: addSource
                text: "Add music to your collection"

                onClicked: DbClient.display_folder_chooser()
            }
        }

        Image {
            source: "../../../data/art/dropshadow-fadeout-left.png"
            anchors.top: sourceHeader.bottom
            anchors.right: dropShadow.left
        }

        Image {
            source: "../../../data/art/dropshadow-fadeout-right.png"
            anchors.top: sourceHeader.bottom
            anchors.left: dropShadow.right
        }

        Image {
            id: dropShadow
            source: "../../../data/art/dropshadow-north.png"
            width: parent.width
            anchors.top: sourceHeader.bottom
            anchors.left: parent.left
        }
    }

    //    Scrollbar {
    //        anchors.right: parent.right
    //        anchors.top: parent.top
    //        anchors.bottom: parent.bottom
    //        anchors.margins: 3
    //        scrollArea: indexes
    //    }
}
